# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import traceback
import uuid

import eventlet.debug
from oslo_config import cfg
from oslo_log import log as logging
from oslo_messaging import target
from oslo_serialization import jsonutils
from oslo_service import service

from murano.common import auth_utils
from murano.common.helpers import token_sanitizer
from murano.common.plugins import extensions_loader
from murano.common import rpc
from murano.dsl import context_manager
from murano.dsl import dsl_exception
from murano.dsl import executor as dsl_executor
from murano.dsl import helpers
from murano.dsl import schema_generator
from murano.dsl import serializer
from murano.engine import execution_session
from murano.engine import package_loader
from murano.engine.system import status_reporter
from murano.engine.system import yaql_functions
from murano.policy import model_policy_enforcer as enforcer

CONF = cfg.CONF

PLUGIN_LOADER = None

LOG = logging.getLogger(__name__)

eventlet.debug.hub_exceptions(False)


# noinspection PyAbstractClass
class EngineService(service.Service):
    def __init__(self):
        super(EngineService, self).__init__()
        self.server = None

    def start(self):
        if not rpc.initialized():
            rpc.init()
        endpoints = [
            TaskProcessingEndpoint(),
            StaticActionEndpoint(),
            SchemaEndpoint()
        ]

        s_target = target.Target('murano', 'tasks', server=str(uuid.uuid4()))
        self.server = rpc.get_server(s_target, endpoints, executor='eventlet')
        self.server.start()
        super(EngineService, self).start()

    def stop(self, graceful=False):
        if self.server:
            self.server.stop()
            if graceful:
                self.server.wait()
        super(EngineService, self).stop()

    def reset(self):
        if self.server:
            self.server.reset()
        super(EngineService, self).reset()


def get_plugin_loader():
    global PLUGIN_LOADER

    if PLUGIN_LOADER is None:
        PLUGIN_LOADER = extensions_loader.PluginLoader()
    return PLUGIN_LOADER


class ContextManager(context_manager.ContextManager):
    def create_root_context(self, runtime_version):
        root_context = super(ContextManager, self).create_root_context(
            runtime_version)
        return helpers.link_contexts(
            root_context, yaql_functions.get_context(runtime_version))

    def create_package_context(self, package):
        context = super(ContextManager, self).create_package_context(
            package)
        if package.name == 'io.murano':
            context = helpers.link_contexts(
                context, yaql_functions.get_restricted_context())
        return context


class SchemaEndpoint(object):
    @classmethod
    def generate_schema(cls, context, *args, **kwargs):
        session = execution_session.ExecutionSession()
        session.token = context['token']
        session.project_id = context['project_id']
        with package_loader.CombinedPackageLoader(session) as pkg_loader:
            return schema_generator.generate_schema(
                pkg_loader, ContextManager(), *args, **kwargs)


class TaskProcessingEndpoint(object):
    @classmethod
    def handle_task(cls, context, task):
        result = cls.execute(task)
        rpc.api().process_result(result, task['id'])

    @staticmethod
    def execute(task):
        s_task = token_sanitizer.TokenSanitizer().sanitize(task)
        LOG.info('Starting processing task: {task_desc}'.format(
            task_desc=jsonutils.dumps(s_task)))

        result = None
        reporter = status_reporter.StatusReporter(task['id'])

        try:
            task_executor = TaskExecutor(task, reporter)
            result = task_executor.execute()
            return result
        finally:
            s_result = token_sanitizer.TokenSanitizer().sanitize(result)
            LOG.info('Finished processing task: {task_desc}'.format(
                task_desc=jsonutils.dumps(s_result)))


class StaticActionEndpoint(object):
    @classmethod
    def call_static_action(cls, context, task):
        s_task = token_sanitizer.TokenSanitizer().sanitize(task)
        LOG.info('Starting execution of static action: '
                 '{task_desc}'.format(task_desc=jsonutils.dumps(s_task)))

        result = None
        reporter = status_reporter.StatusReporter(task['id'])

        try:
            task_executor = StaticActionExecutor(task, reporter)
            result = task_executor.execute()
            return result
        finally:
            LOG.info('Finished execution of static action: '
                     '{task_desc}'.format(task_desc=jsonutils.dumps(result)))


class TaskExecutor(object):
    @property
    def action(self):
        return self._action

    @property
    def session(self):
        return self._session

    @property
    def model(self):
        return self._model

    def __init__(self, task, reporter=None):
        if reporter is None:
            reporter = status_reporter.StatusReporter(task['id'])
        self._action = task.get('action')
        self._model = task['model']
        self._session = execution_session.ExecutionSession()
        self._session.token = task['token']
        self._session.project_id = task['project_id']
        self._session.user_id = task['user_id']
        self._session.environment_owner_project_id = self._model['project_id']
        self._session.environment_owner_user_id = self._model['user_id']
        self._session.system_attributes = self._model.get('SystemData', {})
        self._reporter = reporter

        self._model_policy_enforcer = enforcer.ModelPolicyEnforcer(
            self._session)

    def execute(self):
        try:
            self._create_trust()
        except Exception as e:
            return self.exception_result(e, None, '<system>')

        with package_loader.CombinedPackageLoader(self._session) as pkg_loader:
            pkg_loader.import_fixation_table(
                self._session.system_attributes.get('Packages', {}))
            result = self._execute(pkg_loader)
            self._session.system_attributes[
                'Packages'] = pkg_loader.export_fixation_table()
        self._model['SystemData'] = self._session.system_attributes
        self._model['project_id'] = self._session.environment_owner_project_id
        self._model['user_id'] = self._session.environment_owner_user_id
        result['model'] = self._model

        if (not self._model.get('Objects') and
                not self._model.get('ObjectsCopy')):
            try:
                self._delete_trust()
            except Exception:
                LOG.warning('Cannot delete trust', exc_info=True)

        return result

    def _execute(self, pkg_loader):

        get_plugin_loader().register_in_loader(pkg_loader)
        with dsl_executor.MuranoDslExecutor(
                pkg_loader, ContextManager(), self.session) as executor:
            try:
                obj = executor.load(self.model)
            except Exception as e:
                return self.exception_result(e, None, '<load>')

            if obj is not None:
                try:
                    self._validate_model(obj.object, pkg_loader, executor)
                except Exception as e:
                    return self.exception_result(e, obj, '<validate>')
            try:
                LOG.debug('Invoking pre-cleanup hooks')
                self.session.start()
                executor.object_store.cleanup()
            except Exception as e:
                return self.exception_result(e, obj, '<GC>')
            finally:
                LOG.debug('Invoking post-cleanup hooks')
                self.session.finish()
            self._model['ObjectsCopy'] = \
                copy.deepcopy(self._model.get('Objects'))

            action_result = None
            if self.action:
                try:
                    LOG.debug('Invoking pre-execution hooks')
                    self.session.start()
                    action_result = self._invoke(executor)
                except Exception as e:
                    return self.exception_result(e, obj, self.action['method'])
                finally:
                    LOG.debug('Invoking post-execution hooks')
                    self.session.finish()
                    self._model = executor.finalize(obj)
            try:
                action_result = serializer.serialize(action_result, executor)
            except Exception as e:
                return self.exception_result(e, None, '<result>')

        pkg_loader.compact_fixation_table()
        return {
            'action': {
                'result': action_result,
                'isException': False
            }
        }

    def exception_result(self, exception, root, method_name):
        if isinstance(exception, dsl_exception.MuranoPlException):
            LOG.error('\n' + exception.format(prefix='  '))
            exception_traceback = exception.format()
        else:
            exception_traceback = traceback.format_exc()
            LOG.exception(
                ("Exception %(exc)s occurred"
                 " during invocation of %(method)s"),
                {'exc': exception, 'method': method_name})
        self._reporter.report_error(root, str(exception))

        return {
            'action': {
                'isException': True,
                'result': {
                    'message': str(exception),
                    'details': exception_traceback
                }
            }
        }

    def _validate_model(self, obj, pkg_loader, executor):
        if CONF.engine.enable_model_policy_enforcer:
            if obj is not None:
                with helpers.with_object_store(executor.object_store):
                    self._model_policy_enforcer.modify(obj, pkg_loader)
                    self._model_policy_enforcer.validate(obj.to_dictionary(),
                                                         pkg_loader)

    def _invoke(self, mpl_executor):
        obj = mpl_executor.object_store.get(self.action['object_id'])
        method_name, kwargs = self.action['method'], self.action['args']

        if obj is not None:
            return mpl_executor.run(obj.type, method_name, obj, (), kwargs)

    def _create_trust(self):
        if not CONF.engine.use_trusts:
            return
        trust_id = self._session.system_attributes.get('TrustId')
        if not trust_id:
            trust_id = auth_utils.create_trust(
                self._session.token, self._session.project_id)
            self._session.system_attributes['TrustId'] = trust_id
        self._session.trust_id = trust_id

    def _delete_trust(self):
        trust_id = self._session.trust_id
        if trust_id:
            auth_utils.delete_trust(self._session)
            self._session.system_attributes['TrustId'] = None
            self._session.trust_id = None


class StaticActionExecutor(object):
    @property
    def action(self):
        return self._action

    @property
    def session(self):
        return self._session

    def __init__(self, task, reporter=None):
        if reporter is None:
            reporter = status_reporter.StatusReporter(task['id'])
        self._action = task['action']
        self._session = execution_session.ExecutionSession()
        self._session.token = task['token']
        self._session.project_id = task['project_id']
        self._session.user_id = task['user_id']
        self._reporter = reporter
        self._model_policy_enforcer = enforcer.ModelPolicyEnforcer(
            self._session)

    def execute(self):
        with package_loader.CombinedPackageLoader(self._session) as pkg_loader:
            get_plugin_loader().register_in_loader(pkg_loader)
            executor = dsl_executor.MuranoDslExecutor(pkg_loader,
                                                      ContextManager())
            action_result = self._invoke(executor)
            action_result = serializer.serialize(action_result, executor)
            return action_result

    def _invoke(self, mpl_executor):
        class_name = self.action['class_name']
        pkg_name = self.action['pkg_name']
        class_version = self.action['class_version']
        version_spec = helpers.parse_version_spec(class_version)
        if pkg_name:
            package = mpl_executor.package_loader.load_package(
                pkg_name, version_spec)
        else:
            package = mpl_executor.package_loader.load_class_package(
                class_name, version_spec)
        cls = package.find_class(class_name, search_requirements=False)
        method_name, kwargs = self.action['method'], self.action['args']

        return mpl_executor.run(cls, method_name, None, (), kwargs)



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from murano.api.v1 import request_statistics
from murano.common import policy
from murano.common import wsgi
from murano.db.services import instances


LOG = logging.getLogger(__name__)
API_NAME = 'EnvironmentStatistics'


class Controller(object):
    @request_statistics.stats_count(API_NAME, 'GetAggregated')
    def get_aggregated(self, request, environment_id):
        LOG.debug('EnvironmentStatistics:GetAggregated')
        target = {"environment_id": environment_id}
        policy.check("get_aggregated_statistics", request.context, target)

        # TODO(stanlagun): Check that caller is authorized to access
        #  tenant's statistics

        return instances.InstanceStatsServices.get_aggregated_stats(
            environment_id)

    @request_statistics.stats_count(API_NAME, 'GetForInstance')
    def get_for_instance(self, request, environment_id, instance_id):
        LOG.debug('EnvironmentStatistics:GetForInstance')
        target = {"environment_id": environment_id, "instance_id": instance_id}
        policy.check("get_instance_statistics", request.context, target)

        # TODO(stanlagun): Check that caller is authorized to access
        #  tenant's statistics

        return instances.InstanceStatsServices.get_raw_environment_stats(
            environment_id, instance_id)

    @request_statistics.stats_count(API_NAME, 'GetForEnvironment')
    def get_for_environment(self, request, environment_id):
        LOG.debug('EnvironmentStatistics:GetForEnvironment')
        target = {"environment_id": environment_id}
        policy.check("get_statistics", request.context, target)

        # TODO(stanlagun): Check that caller is authorized to access
        #  tenant's statistics

        return instances.InstanceStatsServices.get_raw_environment_stats(
            environment_id)


def create_resource():
    return wsgi.Resource(Controller())



# Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestConfigProperties(test_case.DslTestCase):
    def test_config_property(self):
        obj = om.Object('ConfigProperties')
        self.package_loader.set_config_value(obj, 'cfgProperty', '987')
        runner = self.new_runner(obj)
        runner.testPropertyValues()
        self.assertEqual(
            [987, 'DEFAULT'],
            self.traces
        )

    def test_config_property_exclusion_from_obect_model(self):
        obj = om.Object('ConfigProperties', cfgProperty=555)
        runner = self.new_runner(obj)
        runner.testPropertyValues()
        self.assertEqual(
            [123, 'DEFAULT'],
            self.traces
        )

    def test_config_affects_default(self):
        obj = om.Object('ConfigProperties')
        self.package_loader.set_config_value(obj, 'normalProperty', 'custom')
        runner = self.new_runner(obj)
        runner.testPropertyValues()
        self.assertEqual(
            [123, 'custom'],
            self.traces
        )

    def test_config_not_affects_in_properties(self):
        obj = om.Object('ConfigProperties', normalProperty='qq')
        self.package_loader.set_config_value(obj, 'normalProperty', 'custom')
        runner = self.new_runner(obj)
        runner.testPropertyValues()
        self.assertEqual(
            [123, 'qq'],
            self.traces
        )



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.dsl import constants
from murano.dsl import dsl_exception
from murano.dsl import exceptions
from murano.dsl import expressions
from murano.dsl import helpers
from murano.dsl import yaql_expression


class CodeBlock(expressions.DslExpression):
    def __init__(self, body):
        body = helpers.list_value(body)
        self.code_block = list(map(expressions.parse_expression, body))

    def execute(self, context):
        for expr in self.code_block:
            if hasattr(expr, 'virtual_instruction'):
                instruction = expr.virtual_instruction
                context[constants.CTX_CURRENT_INSTRUCTION] = instruction

            try:
                expr.execute(context)
            except (dsl_exception.MuranoPlException,
                    exceptions.InternalFlowException):
                raise
            except Exception as ex:
                raise dsl_exception.MuranoPlException.from_python_exception(
                    ex, context)


class MethodBlock(CodeBlock):
    def __init__(self, body, name=None):
        super(MethodBlock, self).__init__(body)
        self._name = name

    def execute(self, context):
        new_context = context.create_child_context()
        new_context[constants.CTX_VARIABLE_SCOPE] = True
        try:
            super(MethodBlock, self).execute(new_context)
        except exceptions.ReturnException as e:
            return e.value
        except exceptions.BreakException:
            raise exceptions.DslInvalidOperationError(
                'Break cannot be used on method level')
        except exceptions.ContinueException:
            raise exceptions.DslInvalidOperationError(
                'Continue cannot be used on method level')
        else:
            return None


class ReturnMacro(expressions.DslExpression):
    def __init__(self, Return):
        self._value = Return

    def execute(self, context):
        raise exceptions.ReturnException(
            helpers.evaluate(self._value, context))


class BreakMacro(expressions.DslExpression):
    def __init__(self, Break):
        if Break:
            raise exceptions.DslSyntaxError('Break cannot have value')

    def execute(self, context):
        raise exceptions.BreakException()


class ContinueMacro(expressions.DslExpression):
    def __init__(self, Continue):
        if Continue:
            raise exceptions.DslSyntaxError('Continue cannot have value')

    def execute(self, context):
        raise exceptions.ContinueException()


class ParallelMacro(CodeBlock):
    def __init__(self, Parallel, Limit=None):
        super(ParallelMacro, self).__init__(Parallel)
        self._limit = Limit or len(self.code_block)

    def execute(self, context):
        if not self.code_block:
            return
        limit = helpers.evaluate(self._limit, context)
        helpers.parallel_select(
            self.code_block,
            lambda expr: expr.execute(context.create_child_context()),
            limit)


class IfMacro(expressions.DslExpression):
    def __init__(self, If, Then, Else=None):
        self._code1 = CodeBlock(Then)
        self._code2 = None if Else is None else CodeBlock(Else)
        self._condition = If

    def execute(self, context):
        if helpers.evaluate(self._condition, context):
            self._code1.execute(context)
        elif self._code2 is not None:
            self._code2.execute(context)


class WhileDoMacro(expressions.DslExpression):
    def __init__(self, While, Do):
        if not isinstance(While, yaql_expression.YaqlExpression):
            raise TypeError()
        self._code = CodeBlock(Do)
        self._condition = While

    def execute(self, context):
        while self._condition(context):
            try:
                self._code.execute(context)
            except exceptions.BreakException:
                break
            except exceptions.ContinueException:
                continue


class ForMacro(expressions.DslExpression):
    def __init__(self, For, In, Do):
        if not isinstance(For, str):
            raise exceptions.DslSyntaxError(
                'For value must be of string type')
        self._code = CodeBlock(Do)
        self._var = For
        self._collection = In

    def execute(self, context):
        collection = helpers.evaluate(self._collection, context)
        for t in collection:
            context[self._var] = t
            try:
                self._code.execute(context)
            except exceptions.BreakException:
                break
            except exceptions.ContinueException:
                continue


class RepeatMacro(expressions.DslExpression):
    def __init__(self, Repeat, Do):
        if not isinstance(Repeat, (int, yaql_expression.YaqlExpression)):
            raise exceptions.DslSyntaxError(
                'Repeat value must be either int or expression')
        self._count = Repeat
        self._code = CodeBlock(Do)

    def execute(self, context):
        count = helpers.evaluate(self._count, context)
        for _ in range(0, count):
            try:
                self._code.execute(context)
            except exceptions.BreakException:
                break
            except exceptions.ContinueException:
                continue


class MatchMacro(expressions.DslExpression):
    def __init__(self, Match, Value, Default=None):
        if not isinstance(Match, dict):
            raise exceptions.DslSyntaxError(
                'Match value must be of dictionary type')
        self._switch = Match
        self._value = Value
        self._default = None if Default is None else CodeBlock(Default)

    def execute(self, context):
        match_value = helpers.evaluate(self._value, context)
        for key, value in self._switch.items():
            if key == match_value:
                CodeBlock(value).execute(context)
                return
        if self._default is not None:
            self._default.execute(context)


class SwitchMacro(expressions.DslExpression):
    def __init__(self, Switch, Default=None):
        if not isinstance(Switch, dict):
            raise exceptions.DslSyntaxError(
                'Switch value must be of dictionary type')
        self._switch = Switch
        self._default = None if Default is None else CodeBlock(Default)

    def execute(self, context):
        matched = False
        for key, value in self._switch.items():
            if helpers.evaluate(key, context):
                matched = True
                CodeBlock(value).execute(context)

        if self._default is not None and not matched:
            self._default.execute(context)


class DoMacro(expressions.DslExpression):
    def __init__(self, Do):
        self._code = CodeBlock(Do)

    def execute(self, context):
        self._code.execute(context)


def register():
    expressions.register_macro(DoMacro)
    expressions.register_macro(ReturnMacro)
    expressions.register_macro(BreakMacro)
    expressions.register_macro(ContinueMacro)
    expressions.register_macro(ParallelMacro)
    expressions.register_macro(IfMacro)
    expressions.register_macro(WhileDoMacro)
    expressions.register_macro(ForMacro)
    expressions.register_macro(RepeatMacro)
    expressions.register_macro(MatchMacro)
    expressions.register_macro(SwitchMacro)



# Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from keystoneclient import exceptions as ks_exceptions
from oslo_config import cfg
from oslo_log import log as logging
import yaml

from murano.common import auth_utils
from murano.common import uuidutils
from murano.db import models
from murano.db.services import sessions
from murano.db import session as db_session
from murano.services import states

CONF = cfg.CONF

LOG = logging.getLogger(__name__)
DEFAULT_NETWORK_TYPES = {
    "nova": 'io.murano.resources.NovaNetwork',
    "neutron": 'io.murano.resources.NeutronNetwork'
}


class EnvironmentServices(object):
    @staticmethod
    def get_environments_by(filters):
        """Returns list of environments

           :param filters: property filters
           :return: Returns list of environments
        """
        unit = db_session.get_session()
        environments = unit.query(models.Environment). \
            filter_by(**filters).all()

        for env in environments:
            env['status'] = EnvironmentServices.get_status(env['id'])

        return environments

    @staticmethod
    def get_status(environment_id):
        """Environment can have one of the following statuses:

         - deploying: there is ongoing deployment for environment
         - deleting: environment is currently being deleted
         - deploy failure: last deployment session has failed
         - delete failure: last delete session has failed
         - pending: there is at least one session with status `open` and no
            errors in previous sessions
         - ready: there are no sessions for environment

        :param environment_id: Id of environment for which we checking status.
        :return: Environment status
        """
        # Deploying: there is at least one valid session with status deploying
        session_list = sessions.SessionServices.get_sessions(environment_id)
        has_opened = False
        for session in session_list:
            if session.state == states.SessionState.DEPLOYING:
                return states.EnvironmentStatus.DEPLOYING
            elif session.state == states.SessionState.DELETING:
                return states.EnvironmentStatus.DELETING
            elif session.state == states.SessionState.DEPLOY_FAILURE:
                return states.EnvironmentStatus.DEPLOY_FAILURE
            elif session.state == states.SessionState.DELETE_FAILURE:
                return states.EnvironmentStatus.DELETE_FAILURE
            elif session.state == states.SessionState.OPENED:
                has_opened = True
            elif session.state == states.SessionState.DEPLOYED:
                break
        if has_opened:
            return states.EnvironmentStatus.PENDING

        return states.EnvironmentStatus.READY

    @staticmethod
    def create(environment_params, context):
        # tagging environment by tenant_id for later checks
        """Creates environment with specified params, in particular - name

           :param environment_params: Dict, e.g. {'name': 'env-name'}
           :param context: request context to get the tenant id and the token
           :return: Created Environment
        """
        objects = {'?': {
            'id': uuidutils.generate_uuid(),
        }}
        network_driver = EnvironmentServices.get_network_driver(context)
        objects.update(environment_params)
        if not objects.get('defaultNetworks'):
            objects['defaultNetworks'] = \
                EnvironmentServices.generate_default_networks(objects['name'],
                                                              network_driver)
        objects['?']['type'] = 'io.murano.Environment'
        objects['?']['metadata'] = {}

        data = {
            'Objects': objects,
            'Attributes': [],
            'project_id': context.tenant,
            'user_id': context.user
        }

        environment_params['tenant_id'] = context.tenant
        environment = models.Environment()
        environment.update(environment_params)

        unit = db_session.get_session()
        with unit.begin():
            unit.add(environment)

        # saving environment as Json to itself
        environment.update({'description': data})
        environment.save(unit)

        return environment

    @staticmethod
    def delete(environment_id, session_id):
        """Deletes environment and notify orchestration engine about deletion

           :param environment_id: Environment that is going to be deleted
           :param session_id: Session Id
        """

        env_description = EnvironmentServices.get_environment_description(
            environment_id, session_id, False)
        env_description['Objects'] = None
        EnvironmentServices.save_environment_description(
            session_id, env_description, False)

    @staticmethod
    def remove(environment_id):
        unit = db_session.get_session()
        environment = unit.query(models.Environment).get(environment_id)
        if environment:
            with unit.begin():
                unit.delete(environment)

    @staticmethod
    def get_environment_description(environment_id, session_id=None,
                                    inner=True):
        """Returns environment description for specified environment.

           If session is specified and not in deploying state function
           returns modified environment description,
           otherwise returns actual environment desc.

           :param environment_id: Environment Id
           :param session_id: Session Id
           :param inner: return contents of environment rather than whole
            Object Model structure
           :return: Environment Description Object
        """
        unit = db_session.get_session()

        if session_id:
            session = unit.query(models.Session).get(session_id)
            if sessions.SessionServices.validate(session):
                if session.state != states.SessionState.DEPLOYED:
                    env_description = session.description
                else:
                    env = unit.query(models.Environment) \
                        .get(session.environment_id)
                    env_description = env.description
            else:
                env = unit.query(models.Environment) \
                    .get(session.environment_id)
                env_description = env.description
        else:
            env = (unit.query(models.Environment).get(environment_id))
            env_description = env.description

        if not inner:
            return env_description
        else:
            return env_description['Objects']

    @staticmethod
    def save_environment_description(session_id, environment, inner=True):
        """Saves environment description to specified session.

           :param session_id: Session Id
           :param environment: Environment Description
           :param inner: save modifications to only content of environment
            rather than whole Object Model structure
        """
        unit = db_session.get_session()
        session = unit.query(models.Session).get(session_id)
        if inner:
            data = session.description.copy()
            data['Objects'] = environment
            session.description = data
        else:
            session.description = environment
        session.save(unit)

    @staticmethod
    def generate_default_networks(env_name, network_driver):
        net_config = CONF.find_file(
            CONF.networking.network_config_file)
        if net_config:
            LOG.debug("Loading network configuration from file")
            with open(net_config) as f:
                data = yaml.safe_load(f)
                return EnvironmentServices._objectify(data, {
                    'ENV': env_name
                })

        network_type = DEFAULT_NETWORK_TYPES[network_driver]
        LOG.debug("Setting '{net_type}' as environment's "
                  "default network".format(net_type=network_type))
        return {
            'environment': {
                '?': {
                    'id': uuidutils.generate_uuid(),
                    'type': network_type
                },
                'name': env_name + '-network'
            },
            'flat': None
        }

    @staticmethod
    def deploy(session, unit, context):
        environment = unit.query(models.Environment).get(
            session.environment_id)

        if (session.description['Objects'] is None and
                'ObjectsCopy' not in session.description):
            EnvironmentServices.remove(session.environment_id)
        else:
            sessions.SessionServices.deploy(
                session, environment, unit, context)

    @staticmethod
    def _objectify(data, replacements):
        if isinstance(data, dict):
            if isinstance(data.get('?'), dict):
                data['?']['id'] = uuidutils.generate_uuid()
            result = {}
            for key, value in data.items():
                result[key] = EnvironmentServices._objectify(
                    value, replacements)
            return result
        elif isinstance(data, list):
            return [EnvironmentServices._objectify(v, replacements)
                    for v in data]
        elif isinstance(data, str):
            for key, value in replacements.items():
                data = data.replace('%' + key + '%', value)
        return data

    @staticmethod
    def get_network_driver(context):
        driver = CONF.networking.driver
        if driver:
            LOG.debug("Will use {} as a network driver".format(driver))
            return driver

        session = auth_utils.get_token_client_session(
            context.auth_token, context.tenant)
        try:
            session.get_endpoint(service_type='network')
        except ks_exceptions.EndpointNotFound:
            LOG.debug("Will use NovaNetwork as a network driver")
            return "nova"
        else:
            LOG.debug("Will use Neutron as a network driver")
            return "neutron"



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.dsl import yaql_integration


# noinspection PyMethodMayBeStatic
class ContextManager(object):
    """Context manager for the MuranoDslExecutor.

    DSL hosting project should subclass this and override methods in order
    to insert yaql function at various scopes. For example it may override
    create_root_context to register its own global yaql functions.
    """

    def create_root_context(self, runtime_version):
        return yaql_integration.create_context(runtime_version)

    def create_package_context(self, package):
        return package.context

    def create_type_context(self, murano_type):
        return murano_type.context

    def create_object_context(self, obj):
        return None



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from murano.db import models
from murano.db import session
from murano.tests.unit import base


class TestModels(base.MuranoWithDBTestCase):
    def test_missing_blob(self):
        """Fake a package with NULL supplier JSON blob to test bug 1342306."""
        con = session.get_session().connection()
        con.execute("INSERT INTO package(id, fully_qualified_name, "
                    "owner_id, name, description, created, updated, type, "
                    "supplier) "
                    "VALUES (1, 'blob.test', 1, 'blob test', 'Desc', "
                    "'2014-07-15 00:00:00', '2014-07-15 00:00:00', "
                    "'Application', NULL)")
        loaded_e = session.get_session().query(models.Package).get(1)
        self.assertIsNone(loaded_e.supplier)



# Copyright (c) 2015 Telefonica I+D.
# Copyright (c) 2016 AT&T Corp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from unittest import mock

from oslo_config import fixture as config_fixture
from oslo_db import exception as db_exc
from oslo_serialization import jsonutils
from oslo_utils import timeutils

from murano.api.v1 import templates
from murano.db import models
import murano.tests.unit.api.base as tb
import murano.tests.unit.utils as test_utils


class TestEnvTemplateApi(tb.ControllerTest, tb.MuranoApiTestCase):

    def setUp(self):
        super(TestEnvTemplateApi, self).setUp()
        self.controller = templates.Controller()
        self.uuids = ['env_template_id', 'other', 'network_id',
                      'environment_id', 'session_id']
        self.mock_uuid = self._stub_uuid(self.uuids)

    def test_list_empty_env_templates(self):
        """Check that with no templates an empty list is returned."""
        self._set_policy_rules(
            {'list_env_templates': '@'}
        )
        self.expect_policy_check('list_env_templates')

        req = self._get('/templates')
        result = req.get_response(self.api)
        self.assertEqual({'templates': []}, jsonutils.loads(result.body))

    def test_create_env_templates(self):
        """Create an template, test template.show()."""
        self._set_policy_rules(
            {'list_env_templates': '@',
             'create_env_template': '@',
             'show_env_template': '@'}
        )
        self.expect_policy_check('create_env_template')

        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now

        expected = {'tenant_id': self.tenant,
                    'id': 'env_template_id',
                    'is_public': False,
                    'name': 'mytemp',
                    'description_text': 'description',
                    'version': 0,
                    'created': datetime.isoformat(fake_now)[:-7],
                    'updated': datetime.isoformat(fake_now)[:-7]}

        body = {'name': 'mytemp', 'description_text': 'description'}
        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)

        self.assertEqual(expected, jsonutils.loads(result.body))

        # Reset the policy expectation
        self.expect_policy_check('list_env_templates')

        req = self._get('/templates')
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)
        self.assertEqual({'templates': [expected]},
                         jsonutils.loads(result.body))

        expected['services'] = []

        self.expect_policy_check('show_env_template',
                                 {'env_template_id': self.uuids[0]})
        req = self._get('/templates/%s' % self.uuids[0])
        result = req.get_response(self.api)
        self.assertEqual(expected, jsonutils.loads(result.body))

    @mock.patch('murano.db.services.environment_templates.EnvTemplateServices.'
                'create')
    def test_create_env_templates_handle_duplicate_exc(self, mock_function):
        """Create an template, test template.show()."""
        self._set_policy_rules(
            {'create_env_template': '@'}
        )
        self.expect_policy_check('create_env_template')
        mock_function.side_effect = db_exc.DBDuplicateEntry

        body = {'name': 'mytemp', 'description_text': 'description'}
        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(409, result.status_code)

    def test_list_public_env_templates(self):
        """Create an template, test templates.public()."""
        self._set_policy_rules(
            {'create_env_template': '@',
             'list_env_templates': '@'}
        )

        self.expect_policy_check('create_env_template')

        body = {'name': 'mytemp2', 'is_public': True}
        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertTrue(jsonutils.loads(result.body)['is_public'])

        self.expect_policy_check('list_env_templates')
        req = self._get('/templates', {'is_public': True})

        result = req.get_response(self.api)
        data = jsonutils.loads(result.body)
        self.assertEqual(1, len(data))
        self.assertTrue(data['templates'][0]['is_public'])

    def test_clone_env_templates(self):
        """Create an template, test templates.public()."""
        self._set_policy_rules(
            {'create_env_template': '@',
             'clone_env_template': '@'}
        )

        self.expect_policy_check('create_env_template')
        body = {'name': 'mytemp2', 'is_public': True}
        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        env_template_id = jsonutils.loads(result.body)['id']
        self.assertTrue(jsonutils.loads(result.body)['is_public'])

        self.expect_policy_check('clone_env_template')
        body = {'name': 'clone', 'is_public': False}
        req = self._post('/templates/%s/clone' % env_template_id,
                         jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertFalse(jsonutils.loads(result.body)['is_public'])
        self.assertEqual('clone', jsonutils.loads(result.body)['name'])

    def test_clone_env_templates_private(self):
        """Create an template, test templates.public()."""
        self._set_policy_rules(
            {'create_env_template': '@',
             'clone_env_template': '@'}
        )

        self.expect_policy_check('create_env_template')
        body = {'name': 'mytemp2', 'is_public': False}
        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        env_template_id = jsonutils.loads(result.body)['id']
        self.assertFalse(jsonutils.loads(result.body)['is_public'])

        self.expect_policy_check('clone_env_template')
        body = {'name': 'clone', 'is_public': False}
        req = self._post('/templates/%s/clone' % env_template_id,
                         jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(result.status_code, 403)

    @mock.patch('murano.db.services.environment_templates.EnvTemplateServices.'
                'clone')
    def test_clone_env_templates_handle_duplicate_exc(self, mock_function):
        """Test whether clone duplication exception is handled correctly."""
        mock_function.side_effect = db_exc.DBDuplicateEntry

        self._set_policy_rules(
            {'create_env_template': '@',
             'clone_env_template': '@'}
        )

        self.expect_policy_check('create_env_template')
        body = {'name': 'mytemp2', 'is_public': True}
        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        env_template_id = jsonutils.loads(result.body)['id']

        self.expect_policy_check('clone_env_template')
        body = {'name': 'clone', 'is_public': False}
        req = self._post('/templates/%s/clone' % env_template_id,
                         jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(409, result.status_code)

    def test_list_public_env_templates_default(self):
        """Test listing public templates when there aren't any

        Create a template; test list public with no
        public templates.
        """
        self._set_policy_rules(
            {'create_env_template': '@',
             'list_env_templates': '@'}
        )

        self.expect_policy_check('create_env_template')
        body = {'name': 'mytemp'}
        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertFalse(jsonutils.loads(result.body)['is_public'])

        self.expect_policy_check('list_env_templates')
        req = self._get('/templates', {'is_public': True})
        result = req.get_response(self.api)

        self.assertFalse(0, len(jsonutils.loads(result.body)))

    def test_list_private_env_templates(self):
        """Test listing private templates

        Create a public template and a private template;
        test list private templates.
        """
        self._set_policy_rules(
            {'create_env_template': '@',
             'list_env_templates': '@'}
        )

        self.expect_policy_check('create_env_template')
        body = {'name': 'mytemp', 'is_public': False}
        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertFalse(jsonutils.loads(result.body)['is_public'])

        self.expect_policy_check('create_env_template')
        body = {'name': 'mytemp1', 'is_public': True}
        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertTrue(jsonutils.loads(result.body)['is_public'])

        self.expect_policy_check('list_env_templates')
        req = self._get('/templates', {'is_public': False})
        result = req.get_response(self.api)
        self.assertEqual(1, len(jsonutils.loads(result.body)['templates']))

    def test_list_env_templates(self):
        """Test listing public templates when there aren't any

        Create a template; test list public with no
        public templates.
        """
        self._set_policy_rules(
            {'create_env_template': '@',
             'list_env_templates': '@'}
        )

        self.expect_policy_check('create_env_template')
        body = {'name': 'mytemp', 'is_public': False}
        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertFalse(jsonutils.loads(result.body)['is_public'])

        self.expect_policy_check('create_env_template')
        body = {'name': 'mytemp1', 'is_public': True}
        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertTrue(jsonutils.loads(result.body)['is_public'])

        self.expect_policy_check('list_env_templates')
        req = self._get('/templates')
        result = req.get_response(self.api)

        self.assertEqual(2, len(jsonutils.loads(result.body)['templates']))

    def test_list_env_templates_with_different_tenant(self):
        """Test listing public template from another tenant

        Create two template in two different tenants;
        test list public template from another tenant.
        """
        self._set_policy_rules(
            {'create_env_template': '@',
             'list_env_templates': '@'}
        )

        self.expect_policy_check('create_env_template')
        body = {'name': 'mytemp', 'is_public': False}
        req = self._post('/templates', jsonutils.dump_as_bytes(body),
                         tenant='first_tenant')
        result = req.get_response(self.api)
        self.assertFalse(jsonutils.loads(result.body)['is_public'])

        self.expect_policy_check('create_env_template')
        body = {'name': 'mytemp1', 'is_public': True}
        req = self._post('/templates', jsonutils.dump_as_bytes(body),
                         tenant='second_tenant')
        result = req.get_response(self.api)
        self.assertTrue(jsonutils.loads(result.body)['is_public'])

        self.expect_policy_check('list_env_templates')
        req = self._get('/templates', tenant='first_tenant')
        result = req.get_response(self.api)

        self.assertEqual(2, len(jsonutils.loads(result.body)['templates']))

    def test_illegal_template_name_create(self):
        """Check that an illegal temp name results in an HTTPClientError."""
        self._set_policy_rules(
            {'list_env_templates': '@',
             'create_env_template': '@',
             'show_env_template': '@'}
        )
        self.expect_policy_check('create_env_template')

        body = {'name': '  '}
        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(400, result.status_code)

    def test_too_long_template_name_create(self):
        """Check that a long template name results in an HTTPBadResquest."""
        self._set_policy_rules(
            {'list_env_templates': '@',
             'create_env_template': '@',
             'show_env_template': '@'}
        )
        self.expect_policy_check('create_env_template')

        body = {'name': 'a' * 256}
        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(400, result.status_code)
        result_msg = result.text.replace('\n', '')
        self.assertIn('Environment template name should be 255 characters '
                      'maximum',
                      result_msg)

    def test_mallformed_body(self):
        """Check that an illegal temp name results in an HTTPClientError."""
        self._set_policy_rules(
            {'create_env_template': '@'}
        )
        self.expect_policy_check('create_env_template')

        body = {'invalid': 'test'}
        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(400, result.status_code)

    def test_missing_env_template(self):
        """Check that a missing env template results in an HTTPNotFound."""
        self._set_policy_rules(
            {'show_env_template': '@'}
        )
        self.expect_policy_check('show_env_template',
                                 {'env_template_id': 'no-such-id'})

        req = self._get('/templates/no-such-id')
        result = req.get_response(self.api)
        self.assertEqual(404, result.status_code)

    def test_update_env_template(self):
        """Check that environment rename works."""
        self._set_policy_rules(
            {'show_env_template': '@',
             'update_env_template': '@'}
        )
        self.expect_policy_check('update_env_template',
                                 {'env_template_id': '12345'})

        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now

        expected = dict(
            id='12345',
            is_public=False,
            name='my-temp',
            version=0,
            created=fake_now,
            updated=fake_now,
            tenant_id=self.tenant,
            description_text='',
            description={
                'name': 'my-temp',
                '?': {'id': '12345'}
            }
        )
        e = models.EnvironmentTemplate(**expected)
        test_utils.save_models(e)

        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now

        del expected['description']
        expected['services'] = []
        expected['name'] = 'renamed_temp'
        expected['updated'] = fake_now

        body = {
            'name': 'renamed_temp'
        }
        req = self._put('/templates/12345', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)

        self.expect_policy_check('show_env_template',
                                 {'env_template_id': '12345'})
        req = self._get('/templates/12345')
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)

        expected['created'] = datetime.isoformat(expected['created'])[:-7]
        expected['updated'] = datetime.isoformat(expected['updated'])[:-7]

        self.assertEqual(expected, jsonutils.loads(result.body))

    def test_update_env_template_with_inappropriate_name(self):
        """Check that environment rename works."""
        self._set_policy_rules(
            {'show_env_template': '@',
             'update_env_template': '@'}
        )
        self.expect_policy_check('update_env_template',
                                 {'env_template_id': '12345'})

        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now

        expected = dict(
            id='12345',
            is_public=False,
            name='my-temp',
            version=0,
            created=fake_now,
            updated=fake_now,
            tenant_id=self.tenant
        )
        e = models.EnvironmentTemplate(**expected)
        test_utils.save_models(e)

        # Attempt to update the environment template with invalid name.
        body = {'name': ''}
        req = self._put('/templates/12345', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(400, result.status_code)
        self.assertIn(b'EnvTemplate body is incorrect', result.body)

        # Verify that the name was not changed.
        self.expect_policy_check('show_env_template',
                                 {'env_template_id': '12345'})
        req = self._get('/templates/12345')
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)
        self.assertEqual(expected['name'],
                         jsonutils.loads(result.body)['name'])

    def test_delete_env_templates(self):
        """Test that environment deletion results in the correct rpc call."""
        self._set_policy_rules(
            {'delete_env_template': '@'}
        )
        self.expect_policy_check(
            'delete_env_template', {'env_template_id': '12345'}
        )

        fake_now = timeutils.utcnow()
        expected = dict(
            id='12345',
            name='my-temp',
            version=0,
            created=fake_now,
            updated=fake_now,
            tenant_id=self.tenant,
            description={
                'name': 'my-temp',
                '?': {'id': '12345'}
            }
        )
        e = models.EnvironmentTemplate(**expected)
        test_utils.save_models(e)

        req = self._delete('/templates/12345')
        result = req.get_response(self.api)

        # Should this be expected behavior?
        self.assertEqual(b'', result.body)
        self.assertEqual(200, result.status_code)

    def test_create_env_templates_with_applications(self):
        """Create an template, test template.show()."""
        self._set_policy_rules(
            {'list_env_templates': '@',
             'create_env_template': '@',
             'show_env_template': '@'}
        )
        self.expect_policy_check('create_env_template')

        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now
        expected = {'tenant_id': self.tenant,
                    'id': self.uuids[0],
                    'is_public': False,
                    'name': 'env_template_name',
                    'description_text': '',
                    'version': 0,
                    'created': datetime.isoformat(fake_now)[:-7],
                    'updated': datetime.isoformat(fake_now)[:-7]}

        services = [
            {
                "instance": {
                    "assignFloatingIp": "true",
                    "keyname": "mykeyname",
                    "image": "cloud-fedora-v3",
                    "flavor": "m1.medium",
                    "?": {
                        "type": "io.murano.resources.Linux",
                        "id": "ef984a74-29a4-45c0-b1dc-2ab9f075732e"
                    }
                },
                "name": "orion",
                "port": "8080",
                "?": {
                    "type": "io.murano.apps.apache.Tomcat",
                    "id": "54cea43d-5970-4c73-b9ac-fea656f3c722"
                }
            }
        ]
        expected['services'] = services

        body = {
            "name": "env_template_name",
            "services": [
                {
                    "instance": {
                        "assignFloatingIp": "true",
                        "keyname": "mykeyname",
                        "image": "cloud-fedora-v3",
                        "flavor": "m1.medium",
                        "?": {
                            "type": "io.murano.resources.Linux",
                            "id": "ef984a74-29a4-45c0-b1dc-2ab9f075732e"
                        }
                    },
                    "name": "orion",
                    "port": "8080",
                    "?": {
                        "type": "io.murano.apps.apache.Tomcat",
                        "id": "54cea43d-5970-4c73-b9ac-fea656f3c722"
                    }
                }
            ]
        }

        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(expected, jsonutils.loads(result.body))

        # Reset the policy expectation
        self.expect_policy_check('list_env_templates')

        req = self._get('/templates')
        result = req.get_response(self.api)
        del expected['services']
        self.assertEqual(200, result.status_code)
        self.assertEqual({'templates': [expected]},
                         jsonutils.loads(result.body))

        # Reset the policy expectation
        self.expect_policy_check('show_env_template',
                                 {'env_template_id': self.uuids[0]})
        expected['services'] = services
        req = self._get('/templates/%s' % self.uuids[0])
        result = req.get_response(self.api)
        self.assertEqual(expected, jsonutils.loads(result.body))

    def test_add_application_to_template(self):
        """Create an template, test template.show()."""
        self._set_policy_rules(
            {'create_env_template': '@',
             'add_application': '@'}
        )
        self.expect_policy_check('create_env_template')

        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now
        services = [
            {
                "instance": {
                    "assignFloatingIp": "true",
                    "keyname": "mykeyname",
                    "image": "cloud-fedora-v3",
                    "flavor": "m1.medium",
                    "?": {
                        "type": "io.murano.resources.Linux",
                        "id": "ef984a74-29a4-45c0-b1dc-2ab9f075732e"
                    }
                },
                "name": "orion",
                "port": "8080",
                "?": {
                    "type": "io.murano.apps.apache.Tomcat",
                    "id": "54cea43d-5970-4c73-b9ac-fea656f3c722"
                }
            }
        ]

        body = {
            "name": "template_name",
        }

        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)

        body = services
        req = self._post('/templates/%s/services' % self.uuids[0],
                         jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)

        self.assertEqual(200, result.status_code)
        self.assertEqual(services, jsonutils.loads(result.body))
        req = self._get('/templates/%s/services' % self.uuids[0])
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)
        self.assertEqual(1, len(jsonutils.loads(result.body)))

        service_no_instance = [
            {
                "instance": "ef984a74-29a4-45c0-b1dc-2ab9f075732e",
                "name": "tomcat",
                "port": "8080",
                "?": {
                    "type": "io.murano.apps.apache.Tomcat",
                    "id": "54cea43d-5970-4c73-b9ac-fea656f3c722"
                }
            }
        ]

        req = self._post('/templates/%s/services' % self.uuids[0],
                         jsonutils.dump_as_bytes(service_no_instance))
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)

        req = self._get('/templates/%s/services' % self.uuids[0])
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)
        self.assertEqual(2, len(jsonutils.loads(result.body)))

    def test_delete_application_in_template(self):
        """Create an template, test template.show()."""
        self._set_policy_rules(
            {'create_env_template': '@',
             'delete_env_application': '@'}
        )
        self.expect_policy_check('create_env_template')

        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now

        body = {
            "name": "mytemplate",
            "services": [
                {
                    "name": "tomcat",
                    "port": "8080",
                    "?": {
                        "type": "io.murano.apps.apache.Tomcat",
                        "id": "54cea43d-5970-4c73-b9ac-fea656f3c722"
                    }
                }
            ]
        }

        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)
        self.assertEqual(1, len(jsonutils.loads(result.body)['services']))

        req = self._get('/templates/%s/services' % self.uuids[0])
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)
        self.assertEqual(1, len(jsonutils.loads(result.body)))

        service_id = '54cea43d-5970-4c73-b9ac-fea656f3c722'
        req = self._get('/templates/' + self.uuids[0] +
                        '/services/' + service_id)
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)

        req = self._delete('/templates/' + self.uuids[0] +
                           '/services/' + service_id)
        result = req.get_response(self.api)

        self.assertEqual(200, result.status_code)
        self.assertEqual(0, len(jsonutils.loads(result.body)['services']))

        req = self._get('/templates/' + self.uuids[0] +
                        '/services/' + service_id)
        result = req.get_response(self.api)
        self.assertEqual(404, result.status_code)

    def test_create_environment(self):
        """Test that environment is created, session configured."""

        self.fixture = self.useFixture(config_fixture.Config())
        self.fixture.conf(args=[])

        self._set_policy_rules(
            {'create_env_template': '@',
             'create_environment': '@'}
        )

        self._create_env_template_no_service()
        body_env = {'name': 'my_template'}

        self.expect_policy_check('create_environment',
                                 {'env_template_id': self.uuids[0]})
        req = self._post('/templates/%s/create-environment' %
                         self.uuids[0], jsonutils.dump_as_bytes(body_env))
        session_result = req.get_response(self.api)
        self.assertEqual(200, session_result.status_code)
        self.assertIsNotNone(session_result)
        body_returned = jsonutils.loads(session_result.body)
        self.assertEqual(self.uuids[4], body_returned['session_id'])
        self.assertEqual(self.uuids[3], body_returned['environment_id'])

    @mock.patch('murano.db.services.environments.EnvironmentServices.create')
    def test_create_environment_handle_duplicate_exc(self, mock_function):
        """Test that duplicate entry exception is correctly handled."""
        mock_function.side_effect = db_exc.DBDuplicateEntry

        self.fixture = self.useFixture(config_fixture.Config())
        self.fixture.conf(args=[])

        self._set_policy_rules(
            {'create_env_template': '@',
             'create_environment': '@'}
        )

        self._create_env_template_no_service()
        body_env = {'name': 'my_template'}

        self.expect_policy_check('create_environment',
                                 {'env_template_id': self.uuids[0]})
        req = self._post('/templates/%s/create-environment' %
                         self.uuids[0], jsonutils.dump_as_bytes(body_env))
        session_result = req.get_response(self.api)
        self.assertEqual(409, session_result.status_code)

    def test_create_env_with_template_and_services(self):
        """Test env and session creation with services

        Test that environment is created and session with template
        with services.
        """
        self.fixture = self.useFixture(config_fixture.Config())
        self.fixture.conf(args=[])
        self._set_policy_rules(
            {'create_env_template': '@',
             'create_environment': '@'}
        )
        self._create_env_template_services()

        self.expect_policy_check('create_environment',
                                 {'env_template_id': self.uuids[0]})
        body = {'name': 'my_template'}
        req = self._post('/templates/%s/create-environment' %
                         self.uuids[0], jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)

        self.assertIsNotNone(result)
        self.assertEqual(200, result.status_code)
        body_returned = jsonutils.loads(result.body)
        self.assertEqual(self.uuids[4], body_returned['session_id'])
        self.assertEqual(self.uuids[3], body_returned['environment_id'])

    def test_create_env_with_template_no_services(self):
        """Test env and session creation without services

        Test that environment is created and session with template
        without services.
        """
        self.fixture = self.useFixture(config_fixture.Config())
        self.fixture.conf(args=[])
        self._set_policy_rules(
            {'create_env_template': '@',
             'create_environment': '@'}
        )
        self._create_env_template_no_service()

        self.expect_policy_check('create_environment',
                                 {'env_template_id': self.uuids[0]})
        body = {'name': 'my_template'}

        req = self._post('/templates/%s/create-environment' %
                         self.uuids[0], jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertIsNotNone(result)
        self.assertEqual(200, result.status_code)
        body_returned = jsonutils.loads(result.body)
        self.assertEqual(self.uuids[4], body_returned['session_id'])
        self.assertEqual(self.uuids[3], body_returned['environment_id'])

    def test_update_service_in_template(self):
        """Test the service is updated in the environment template"""
        self.fixture = self.useFixture(config_fixture.Config())
        self.fixture.conf(args=[])
        self._set_policy_rules(
            {'create_env_template': '@',
             'update_service_env_template': '@'}
        )
        updated_env = "UPDATED_ENV"
        env_template = self._create_env_template_services()
        self.expect_policy_check('update_service_env_template')
        env_template["name"] = updated_env

        req = self._put('/templates/{0}/services/{1}'.
                        format(self.uuids[0], "service_id"),
                        jsonutils.dump_as_bytes(env_template))
        result = req.get_response(self.api)
        self.assertIsNotNone(result)
        self.assertEqual(200, result.status_code)
        body_returned = jsonutils.loads(result.body)
        self.assertEqual(updated_env, body_returned['name'])

    def test_mallformed_env_body(self):
        """Check that an illegal temp name results in an HTTPClientError."""
        self._set_policy_rules(
            {'create_env_template': '@',
             'create_environment': '@'}
        )
        self._create_env_template_no_service()

        self.expect_policy_check('create_environment',
                                 {'env_template_id': self.uuids[0]})
        body = {'invalid': 'test'}
        req = self._post('/templates/%s/create-environment' %
                         self.uuids[0], jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(400, result.status_code)

    def test_delete_notexisting_service(self):
        """Check deleting a not existing service, return a 404 error."""
        self._set_policy_rules(
            {'create_env_template': '@',
             'delete_env_application': '@'}
        )
        self.expect_policy_check('create_env_template')

        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now

        body = {
            "name": "mytemplate",
            "services": [
                {
                    "name": "tomcat",
                    "port": "8080",
                    "?": {
                        "type": "io.murano.apps.apache.Tomcat",
                        "id": "ID1"
                    }
                }
            ]
        }

        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)
        self.assertEqual(1, len(jsonutils.loads(result.body)['services']))

        req = self._delete('/templates/{0}/services/{1}'.format(self.uuids[0],
                                                                "NO_EXIST"))
        result = req.get_response(self.api)
        self.assertEqual(404, result.status_code)

    @mock.patch('murano.db.services.environment_templates.EnvTemplateServices.'
                'get_env_template')
    def test_validate_request_handle_forbidden_exc(self, mock_function):
        """Test whether forbidden exception is thrown with different tenant."""
        self._set_policy_rules(
            {'create_env_template': '@',
             'show_env_template': '@',
             'show_env_template': '@'}
        )

        # If is_public is False, then exception should be thrown.
        mock_env_template = mock.MagicMock(is_public=False, tenant_id=-1)
        mock_function.return_value = mock_env_template

        self._create_env_template_no_service()
        self.expect_policy_check('show_env_template',
                                 {'env_template_id': self.uuids[0]})
        req = self._get('/templates/%s' % self.uuids[0])
        result = req.get_response(self.api)
        self.assertEqual(result.status_code, 403)

        # If is_public is True, then no exception should be thrown.
        mock_env_template = mock.MagicMock(is_public=True, tenant_id=-1)
        mock_function.return_value = mock_env_template

        self.expect_policy_check('show_env_template',
                                 {'env_template_id': self.uuids[0]})
        req = self._get('/templates/%s' % self.uuids[0])
        result = req.get_response(self.api)
        self.assertEqual(result.status_code, 200)

    def test_create_env_notexisting_templatebody(self):
        """Check that an illegal temp name results in an HTTPClientError."""
        self._set_policy_rules(
            {'create_environment': '@'}
        )
        env_template_id = 'noexisting'
        self.expect_policy_check('create_environment',
                                 {'env_template_id': env_template_id})

        body = {'name': 'test'}
        req = self._post('/templates/%s/create-environment'
                         % env_template_id, jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        self.assertEqual(404, result.status_code)

    def _create_env_template_no_service(self):
        self.expect_policy_check('create_env_template')
        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now

        req = self._post('/templates',
                         jsonutils.dump_as_bytes({'name': 'name'}))
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)

    def _create_env_template_services(self):
        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now

        self.expect_policy_check('create_env_template')

        fake_now = timeutils.utcnow()
        timeutils.utcnow.override_time = fake_now
        body = {
            "name": "env_template_name",
            "services": [
                {
                    "instance": {
                        "assignFloatingIp": "true",
                        "keyname": "mykeyname",
                        "image": "cloud-fedora-v3",
                        "flavor": "m1.medium",
                        "?": {
                            "type": "io.murano.resources.Linux",
                            "id": "ef984a74-29a4-45c0-b1dc-2ab9f075732e"
                        }
                    },
                    "name": "orion",
                    "port": "8080",
                    "?": {
                        "type": "io.murano.apps.apache.Tomcat",
                        "id": "service_id"
                    }
                }
            ]
        }

        req = self._post('/templates', jsonutils.dump_as_bytes(body))
        result = req.get_response(self.api)
        return result.json



# Copyright 2010-2011 OpenStack Foundation
# Copyright 2012-2013 IBM Corp.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
#
# Ripped off from Nova's test_migrations.py
# The only difference between Nova and this code is usage of alembic instead
# of sqlalchemy migrations.
#
# There is an ongoing work to extact similar code to oslo incubator. Once it is
# extracted we'll be able to remove this file and use oslo.

import io
import os

from alembic import command
from alembic import config as alembic_config
from alembic import migration
from alembic import script as alembic_script
from oslo_config import cfg
from oslo_log import log as logging

import murano.db.migration

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class BaseWalkMigrationTestCase(object):

    ALEMBIC_CONFIG = alembic_config.Config(
        os.path.join(os.path.dirname(murano.db.migration.__file__),
                     'alembic.ini')
    )

    ALEMBIC_CONFIG.murano_config = CONF

    def _configure(self, engine):
        """Repo and database configuration

        For each type of repository we should do some configuration steps.
        For migrate_repo we should set our database under version control.
        For alembic we should configure database settings. For this goal we
        should use oslo.config and openstack.commom.db.sqlalchemy.session with
        database functionality (reset default settings and session cleanup).
        """
        CONF.set_override('connection', str(engine.url), group='database')

    def _alembic_command(self, alembic_command, engine, *args, **kwargs):
        """Redefines alembic command data return setting

        Most of alembic command return data into output.
        We should redefine this setting for getting info.
        """
        self.ALEMBIC_CONFIG.stdout = buf = io.StringIO()
        CONF.set_override('connection', str(engine.url), group='database')
        getattr(command, alembic_command)(*args, **kwargs)
        res = buf.getvalue().strip()
        LOG.debug('Alembic command `{command}` returns: {result}'.format(
            command=alembic_command, result=res))
        return res

    def _up_and_down_versions(self):
        """Store tuple of versions for successful testing

        Since alembic version has a random algorithm of generation
        (SA-migrate has an ordered autoincrement naming) we should store
        a tuple of versions (version for upgrade and version for downgrade)
        for successful testing of migrations in up>down>up mode.
        """

        env = alembic_script.ScriptDirectory.from_config(self.ALEMBIC_CONFIG)
        versions = []
        for rev in env.walk_revisions():
            versions.append((rev.revision, rev.down_revision or '-1'))

        versions.reverse()
        return versions

    def walk_versions(self, engine=None, snake_walk=False, downgrade=True):
        # Determine latest version script from the repo, then
        # upgrade from 1 through to the latest, with no data
        # in the databases. This just checks that the schema itself
        # upgrades successfully.

        self._configure(engine)
        up_and_down_versions = self._up_and_down_versions()
        for ver_up, ver_down in up_and_down_versions:
            # upgrade -> downgrade -> upgrade
            self._migrate_up(engine, ver_up, with_data=True)
            if snake_walk:
                downgraded = self._migrate_down(engine,
                                                ver_down,
                                                with_data=True,
                                                next_version=ver_up)
                if downgraded:
                    self._migrate_up(engine, ver_up)

        if downgrade:
            # Now walk it back down to 0 from the latest, testing
            # the downgrade paths.
            up_and_down_versions.reverse()
            for ver_up, ver_down in up_and_down_versions:
                # downgrade -> upgrade -> downgrade
                downgraded = self._migrate_down(engine,
                                                ver_down, next_version=ver_up)

                if snake_walk and downgraded:
                    self._migrate_up(engine, ver_up)
                    self._migrate_down(engine, ver_down, next_version=ver_up)

    def _get_version_from_db(self, engine):
        """Fetches latest version of migrate repo from database

        For each type of migrate repo, the latest version from db
        will be returned.
        """
        conn = engine.connect()
        try:
            context = migration.MigrationContext.configure(conn)
            version = context.get_current_revision() or '-1'
        finally:
            conn.close()
        return version

    def _migrate(self, engine, version, cmd):
        """Base method for manipulation with migrate repo

        It will upgrade or downgrade the actual database.
        """

        self._alembic_command(cmd, engine, self.ALEMBIC_CONFIG, version)

    def _migrate_down(self, engine, version, with_data=False,
                      next_version=None):
        try:
            self._migrate(engine, version, 'downgrade')
        except NotImplementedError:
            # NOTE(sirp): some migrations, namely release-level
            # migrations, don't support a downgrade.
            return False
        self.assertEqual(version, self._get_version_from_db(engine))

        # NOTE(sirp): `version` is what we're downgrading to (i.e. the 'target'
        # version). So if we have any downgrade checks, they need to be run for
        # the previous (higher numbered) migration.
        if with_data:
            post_downgrade = getattr(
                self, "_post_downgrade_%s" % next_version, None)
            if post_downgrade:
                post_downgrade(engine)

        return True

    def _migrate_up(self, engine, version, with_data=False):
        """Migrate up to a new version of the db.

        We allow for data insertion and post checks at every
        migration version with special _pre_upgrade_### and
        _check_### functions in the main test.
        """
        # NOTE(sdague): try block is here because it's impossible to debug
        # where a failed data migration happens otherwise
        check_version = version
        try:
            if with_data:
                data = None
                pre_upgrade = getattr(
                    self, "_pre_upgrade_%s" % check_version, None)
                if pre_upgrade:
                    data = pre_upgrade(engine)
            self._migrate(engine, version, 'upgrade')
            self.assertEqual(version, self._get_version_from_db(engine))
            if with_data:
                check = getattr(self, "_check_%s" % check_version, None)
                if check:
                    check(engine, data)
        except Exception:
            LOG.error("Failed to migrate to version {ver} on engine {eng}"
                      .format(ver=version, eng=engine))
            raise



#    Copyright (c) 2016 AT&T
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from murano.engine.system import agent_listener
from murano.tests.unit import base


class TestExecutionPlan(base.MuranoTestCase):
    def setUp(self):
        super(TestExecutionPlan, self).setUp()
        self.override_config("disable_murano_agent", False, "engine")
        self.agent = agent_listener.AgentListener("test")
        self.addCleanup(mock.patch.stopall)

    def test_agent_ready(self):
        self.assertEqual({}, self.agent._subscriptions)
        results_queue = str('-execution-results-test')
        self.assertEqual(results_queue, self.agent._results_queue)
        self.assertTrue(self.agent._enabled)
        self.assertIsNone(self.agent._receive_thread)

    def test_queue_name(self):
        self.assertEqual(self.agent._results_queue, self.agent.queue_name())

    @mock.patch("murano.engine.system.agent_listener.dsl.get_this")
    @mock.patch("murano.engine.system."
                "agent_listener.dsl.get_execution_session")
    def test_subscribe_unsubscribe(self, execution_session, mock_this):
        self.agent.subscribe('msg_id', 'event')
        self.assertIn('msg_id', self.agent._subscriptions)
        self.agent.unsubscribe('msg_id')
        self.assertNotIn('msg_id', self.agent._subscriptions)
        self.assertTrue(execution_session.called)
        self.assertTrue(mock_this.called)



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from webob import exc

from murano.common import policy
from murano.common import wsgi
from murano.db.services import environments as envs
from murano.db.services import sessions
from murano.db import session as db_session
from murano.services import actions
from murano.services import states
from murano.utils import verify_env


LOG = logging.getLogger(__name__)


class Controller(object):

    @verify_env
    def execute(self, request, environment_id, action_id, body):
        policy.check("execute_action", request.context, {})

        LOG.debug('Action:Execute <ActionId: {0}>'.format(action_id))

        unit = db_session.get_session()

        # no new session can be opened if environment has deploying status
        env_status = envs.EnvironmentServices.get_status(environment_id)
        if env_status in (states.EnvironmentStatus.DEPLOYING,
                          states.EnvironmentStatus.DELETING):
            LOG.warning('Could not open session for environment '
                        '<EnvId: {id}>, environment has deploying '
                        'or deleting status.'.format(id=environment_id))
            raise exc.HTTPForbidden()

        user_id = request.context.user
        session = sessions.SessionServices.create(environment_id, user_id)

        if not sessions.SessionServices.validate(session):
            LOG.error('Session <SessionId {id}> '
                      'is invalid'.format(id=session.id))
            raise exc.HTTPForbidden()

        task_id = actions.ActionServices.execute(
            action_id, session, unit, request.context, body or {})
        return {'task_id': task_id}

    @verify_env
    def get_result(self, request, environment_id, task_id):
        policy.check("execute_action", request.context, {})

        LOG.debug('Action:GetResult <TaskId: {id}>'.format(id=task_id))

        unit = db_session.get_session()
        result = actions.ActionServices.get_result(environment_id, task_id,
                                                   unit)

        if result is not None:
            return result
        msg = ('Result for task with environment_id: {env_id} and task_id: '
               '{task_id} was not found.'.format(env_id=environment_id,
                                                 task_id=task_id))
        LOG.error(msg)
        raise exc.HTTPNotFound(msg)


def create_resource():
    return wsgi.Resource(Controller())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db import exception
import sqlalchemy

from murano.db import cfapi_models as models
from murano.db import session as db_session


def set_tenant_for_org(cf_org_id, tenant):
    """Store tenant-org link to db"""
    unit = db_session.get_session()
    try:
        with unit.begin():
            org = models.CFOrganization()
            org.id = cf_org_id
            org.tenant = tenant
            unit.add(org)
    except exception.DBDuplicateEntry:
        unit.execute(sqlalchemy.update(models.CFOrganization).where(
            models.CFOrganization.id == cf_org_id).values(
                tenant=tenant))


def set_environment_for_space(cf_space_id, environment_id):
    """Store env-space link to db"""
    unit = db_session.get_session()
    try:
        with unit.begin():
            space = models.CFSpace()
            space.id = cf_space_id
            space.environment_id = environment_id
            unit.add(space)
    except exception.DBDuplicateEntry:
        unit.execute(sqlalchemy.update(models.CFSpace).where(
            models.CFSpace.id == cf_space_id).values(
                environment_id=environment_id))


def set_instance_for_service(instance_id, service_id, environment_id, tenant):
    """Store env-space link to db"""
    unit = db_session.get_session()
    try:
        with unit.begin():
            connection = models.CFServiceInstance()
            connection.id = instance_id
            connection.service_id = service_id
            connection.environment_id = environment_id
            connection.tenant = tenant
            unit.add(connection)
    except exception.DBDuplicateEntry:
        unit.execute(sqlalchemy.update(models.CFServiceInstance).where(
            models.CFServiceInstance.id == instance_id).values(
                environment_id=environment_id))


def get_environment_for_space(cf_space_id):
    """Take env id related to space from db"""
    unit = db_session.get_session()
    connection = unit.query(models.CFSpace).get(cf_space_id)
    return connection.environment_id


def get_tenant_for_org(cf_org_id):
    """Take tenant id related to org from db"""
    unit = db_session.get_session()
    connection = unit.query(models.CFOrganization).get(cf_org_id)
    return connection.tenant


def get_service_for_instance(instance_id):
    unit = db_session.get_session()
    connection = unit.query(models.CFServiceInstance).get(instance_id)
    return connection


def delete_environment_from_space(environment_id):
    unit = db_session.get_session()
    unit.query(models.CFSpace).filter(
        models.CFSpace.environment_id == environment_id).delete()



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import itertools
import traceback

import eventlet
import eventlet.event
from oslo_log import log as logging
from yaql.language import exceptions as yaql_exceptions
from yaql.language import specs
from yaql.language import utils

from murano.dsl import attribute_store
from murano.dsl import constants
from murano.dsl import dsl
from murano.dsl import dsl_exception
from murano.dsl import dsl_types
from murano.dsl import exceptions as dsl_exceptions
from murano.dsl import helpers
from murano.dsl import object_store
from murano.dsl.principal_objects import stack_trace
from murano.dsl import serializer
from murano.dsl import yaql_integration

LOG = logging.getLogger(__name__)


class MuranoDslExecutor(object):
    def __init__(self, package_loader, context_manager, session=None):
        self._package_loader = package_loader
        self._context_manager = context_manager
        self._session = session
        self._attribute_store = attribute_store.AttributeStore()
        self._object_store = object_store.ObjectStore(self)
        self._locks = {}
        self._root_context_cache = {}
        self._static_properties = {}

    @property
    def object_store(self):
        return self._object_store

    @property
    def execution_session(self):
        return self._session

    @property
    def attribute_store(self):
        return self._attribute_store

    @property
    def package_loader(self):
        return self._package_loader

    @property
    def context_manager(self):
        return self._context_manager

    def invoke_method(self, method, this, context, args, kwargs,
                      skip_stub=False, invoke_action=True):
        if isinstance(this, dsl.MuranoObjectInterface):
            this = this.object
        kwargs = utils.filter_parameters_dict(kwargs)
        runtime_version = method.declaring_type.package.runtime_version
        yaql_engine = yaql_integration.choose_yaql_engine(runtime_version)
        if context is None or not skip_stub:
            actions_only = (context is None and not method.name.startswith('.')
                            and invoke_action)
            method_context = self.create_method_context(
                self.create_object_context(this, context), method)
            method_context[constants.CTX_SKIP_FRAME] = True
            method_context[constants.CTX_ACTIONS_ONLY] = actions_only

            stub = method.static_stub if isinstance(
                this, dsl_types.MuranoType) else method.instance_stub
            if stub is None:
                raise ValueError(
                    'Method {0} cannot be called on receiver {1}'.format(
                        method, this))

            real_this = this.real_this if isinstance(
                this, dsl_types.MuranoObject) else this.get_reference()
            return stub(yaql_engine, method_context, real_this)(
                *args, **kwargs)

        if context[constants.CTX_ACTIONS_ONLY] and not method.is_action:
            raise dsl_exceptions.MethodNotExposed(
                '{0} is not an action'.format(method.name))

        if method.is_static:
            obj_context = self.create_object_context(
                method.declaring_type, context)
        else:
            obj_context = self.create_object_context(this, context)
        context = self.create_method_context(obj_context, method)

        if isinstance(this, dsl_types.MuranoObject):
            if this.destroyed:
                raise dsl_exceptions.ObjectDestroyedError(this)
            this = this.real_this

        if method.arguments_scheme is not None:
            args, kwargs = self._canonize_parameters(
                method.arguments_scheme, args, kwargs, method.name, this)

        this_lock = this
        arg_values_for_lock = {}
        method_meta = [m for m in method.get_meta(context)
                       if m.type.name == ('io.murano.metadata.'
                                          'engine.Synchronize')]
        if method_meta:
            method_meta = method_meta[0]

        if method_meta:
            if not method_meta.get_property('onThis', context):
                this_lock = None
            for arg_name in method_meta.get_property('onArgs', context):
                arg_val = kwargs.get(arg_name)
                if arg_val is not None:
                    arg_values_for_lock[arg_name] = arg_val

        arg_values_for_lock = utils.filter_parameters_dict(arg_values_for_lock)

        with self._acquire_method_lock(method, this_lock, arg_values_for_lock):
            for i, arg in enumerate(args, 2):
                context[str(i)] = arg
            for key, value in kwargs.items():
                context[key] = value

            def call():
                if isinstance(method.body, specs.FunctionDefinition):
                    if isinstance(this, dsl_types.MuranoType):
                        native_this = this.get_reference()
                    else:
                        native_this = dsl.MuranoObjectInterface(this.cast(
                            method.declaring_type))
                    return method.body(
                        yaql_engine, context, native_this)(*args, **kwargs)
                else:
                    context[constants.CTX_NAMES_SCOPE] = \
                        method.declaring_type
                    return (None if method.body is None
                            else method.body.execute(context))

            if (not isinstance(method.body, specs.FunctionDefinition) or
                    not method.body.meta.get(constants.META_NO_TRACE)):
                with self._log_method(context, args, kwargs) as log:
                    result = call()
                    log(result)
                    return result
            else:
                return call()

    @contextlib.contextmanager
    def _acquire_method_lock(self, method, this, arg_val_dict):
        if this is None:
            if not arg_val_dict:
                # if neither "this" nor argument values are set then no
                # locking is needed
                key = None
            else:
                # if only the argument values are passed then find the lock
                # list only by the method
                key = (None, id(method))
        else:
            if method.is_static:
                # find the lock list by the type and method
                key = (id(method.declaring_type), id(method))
            else:
                # find the lock list by the object and method
                key = (this.object_id, id(method))
        thread_id = helpers.get_current_thread_id()
        while True:
            event, event_owner = None, None
            if key is None:  # no locking needed
                break

            lock_list = self._locks.setdefault(key, [])
            # lock list contains a list of tuples:
            # first item of each tuple is a dict with the values of locking
            # arguments (it is used for argument values comparison),
            # second item is an event to wait on,
            # third one is the owner thread id

            # If this lock list is empty it means no locks on this object and
            # method at all.
            for arg_vals, l_event, l_event_owner in lock_list:
                if arg_vals == arg_val_dict:
                    event = l_event
                    event_owner = l_event_owner
                    break

            if event:
                if event_owner == thread_id:
                    # this means a re-entrant lock: the tuple with the same
                    # value of the first element exists in the list, but it was
                    # acquired by the same green thread. We may proceed with
                    # the call in this case
                    event = None
                    break
                else:
                    event.wait()
            else:
                # this means either the lock list was empty or didn't contain a
                # tuple with the first element equal to arg_val_dict.
                # Then let's acquire a lock, i.e. create a new tuple and place
                # it into the list
                event = eventlet.event.Event()
                event_owner = thread_id
                lock_list.append((arg_val_dict, event, event_owner))
                break
        try:
            yield
        finally:
            if event is not None:
                lock_list.remove((arg_val_dict, event, event_owner))
                if len(lock_list) == 0:
                    del self._locks[key]
                event.send()

    @contextlib.contextmanager
    def _log_method(self, context, args, kwargs):
        method = helpers.get_current_method(context)
        param_gen = itertools.chain(
            (str(arg) for arg in args),
            (u'{0} => {1}'.format(name, value)
             for name, value in kwargs.items()))
        params_str = u', '.join(param_gen)
        method_name = '::'.join((method.declaring_type.name, method.name))
        thread_id = helpers.get_current_thread_id()
        caller_str = ''
        caller_ctx = helpers.get_caller_context(context)
        if caller_ctx is not None:
            frame = stack_trace.compose_stack_frame(caller_ctx)
            if frame['location']:
                caller_str = ' called from ' + stack_trace.format_frame(frame)

        LOG.trace(u'{thread}: Begin execution {method}({params}){caller}'
                  .format(thread=thread_id, method=method_name,
                          params=params_str, caller=caller_str))
        try:
            def log_result(result):
                LOG.trace(
                    u'{thread}: End execution {method} with result '
                    u'{result}'.format(
                        thread=thread_id, method=method_name, result=result))
            yield log_result
        except Exception as e:
            LOG.trace(
                u'{thread}: End execution {method} with exception '
                u'{exc}'.format(thread=thread_id, method=method_name, exc=e))
            raise

    @staticmethod
    def _canonize_parameters(arguments_scheme, args, kwargs,
                             method_name, receiver):
        arg_names = list(arguments_scheme.keys())
        parameter_values = {}
        varargs_arg = None
        vararg_values = []
        kwargs_arg = None
        kwarg_values = {}
        for name, definition in arguments_scheme.items():
            if definition.usage == dsl_types.MethodArgumentUsages.VarArgs:
                varargs_arg = name
                parameter_values[name] = vararg_values
            elif definition.usage == dsl_types.MethodArgumentUsages.KwArgs:
                kwargs_arg = name
                parameter_values[name] = kwarg_values

        for i, arg in enumerate(args):
            name = None if i >= len(arg_names) else arg_names[i]
            if name is None or name in (varargs_arg, kwargs_arg):
                if varargs_arg:
                    vararg_values.append(arg)
                else:
                    raise yaql_exceptions.NoMatchingMethodException(
                        method_name, receiver)
            else:
                parameter_values[name] = arg

        for name, value in utils.filter_parameters_dict(kwargs).items():
            if name in arguments_scheme and name not in (
                    varargs_arg, kwargs_arg):
                parameter_values[name] = value
            elif kwargs_arg:
                kwarg_values[name] = value
            else:
                raise yaql_exceptions.NoMatchingMethodException(
                    method_name, receiver)
        return tuple(), parameter_values

    def load(self, data):
        with helpers.with_object_store(self.object_store):
            return self._load(data)

    def _load(self, data):
        if not isinstance(data, dict):
            raise TypeError()
        self._attribute_store.load(data.get(constants.DM_ATTRIBUTES) or [])
        model = data.get(constants.DM_OBJECTS)
        if model is None:
            result = None
        else:
            result = self._object_store.load(model, None, keep_ids=True)
        model_copy = data.get(constants.DM_OBJECTS_COPY)
        if model_copy:
            self._object_store.load(model_copy, None, keep_ids=True)
        return dsl.MuranoObjectInterface.create(result)

    def signal_destruction_dependencies(self, *objects):
        if not objects:
            return
        elif len(objects) > 1:
            return helpers.parallel_select(
                objects, self.signal_destruction_dependencies)

        obj = objects[0]
        if obj.destroyed:
            return
        for dependency in obj.destruction_dependencies:
            try:
                handler = dependency['handler']
                if handler:
                    subscriber = dependency['subscriber']
                    if subscriber:
                        subscriber = subscriber()
                    if (subscriber and
                            subscriber.initialized and
                            not subscriber.destroyed):
                        method = subscriber.type.find_single_method(handler)
                        self.invoke_method(
                            method, subscriber, None, [obj], {},
                            invoke_action=False)
            except Exception as e:
                LOG.warning('Muted exception during destruction dependency '
                            'execution in {0}: {1}'.format(obj, e),
                            exc_info=True)
        obj.load_dependencies(None)

    def destroy_objects(self, *objects):
        if not objects:
            return
        elif len(objects) > 1:
            return helpers.parallel_select(
                objects, self.destroy_objects)

        obj = objects[0]
        if obj.destroyed:
            return
        methods = obj.type.find_methods(lambda m: m.name == '.destroy')
        for method in methods:
            try:
                method.invoke(obj, (), {}, None)
            except Exception as e:
                if isinstance(e, dsl_exception.MuranoPlException):
                    tb = e.format(prefix='  ')
                else:
                    tb = traceback.format_exc()
                LOG.warning(
                    'Muted exception during execution of .destroy '
                    'on {0}: {1}'.format(obj, tb), exc_info=True)

    def create_root_context(self, runtime_version):
        context = self._root_context_cache.get(runtime_version)
        if not context:
            context = self.context_manager.create_root_context(runtime_version)
            self._root_context_cache[runtime_version] = context
        return context

    def create_package_context(self, package):
        root_context = self.create_root_context(package.runtime_version)
        context = helpers.link_contexts(
            root_context,
            self.context_manager.create_package_context(package))
        return context

    def create_type_context(self, murano_type, caller_context=None):
        package_context = self.create_package_context(
            murano_type.package)
        context = helpers.link_contexts(
            package_context,
            self.context_manager.create_type_context(
                murano_type)).create_child_context()
        context[constants.CTX_TYPE] = murano_type
        if caller_context:
            context[constants.CTX_NAMES_SCOPE] = caller_context[
                constants.CTX_NAMES_SCOPE]
        return context

    def create_object_context(self, obj, caller_context=None):
        if isinstance(obj, dsl_types.MuranoClass):
            obj_type = obj
            obj = None
        else:
            obj_type = obj.type
        class_context = self.create_type_context(obj_type)
        if obj is not None:
            context = helpers.link_contexts(
                class_context, self.context_manager.create_object_context(
                    obj)).create_child_context()
            context[constants.CTX_THIS] = obj.real_this
            context['this'] = obj.real_this
            context[''] = obj.real_this
        else:
            context = class_context.create_child_context()
            type_ref = obj_type.get_reference()
            context[constants.CTX_THIS] = type_ref
            context['this'] = type_ref
            context[''] = type_ref

        if caller_context is not None:
            caller = caller_context
            while caller is not None and caller[constants.CTX_SKIP_FRAME]:
                caller = caller[constants.CTX_CALLER_CONTEXT]
            context[constants.CTX_NAMES_SCOPE] = caller_context[
                constants.CTX_NAMES_SCOPE]
            context[constants.CTX_CALLER_CONTEXT] = caller
            context[constants.CTX_ALLOW_PROPERTY_WRITES] = caller_context[
                constants.CTX_ALLOW_PROPERTY_WRITES]
        else:
            context[constants.CTX_NAMES_SCOPE] = obj_type
        return context

    @staticmethod
    def create_method_context(object_context, method):
        context = object_context.create_child_context()
        context[constants.CTX_CURRENT_METHOD] = method
        return context

    def run(self, cls, method_name, this, args, kwargs):
        with helpers.with_object_store(self.object_store):
            return cls.invoke(method_name, this, args, kwargs)

    def get_static_property(self, murano_type, name, context):
        prop = murano_type.find_static_property(name)
        cls = prop.declaring_type
        value = self._static_properties.get(prop, prop.default)
        return prop.transform(value, cls, None, context)

    def set_static_property(self, murano_type, name, value,
                            context, dry_run=False):
        prop = murano_type.find_static_property(name)
        cls = prop.declaring_type
        value = prop.transform(value, cls, None, context)
        if not dry_run:
            self._static_properties[prop] = prop.finalize(
                value, cls, context)

    def finalize(self, model_root=None):
        # NOTE(ksnihyr): should be no-except
        try:
            if model_root:
                used_objects = serializer.collect_objects(model_root)
                self.object_store.prepare_finalize(used_objects)
                model = serializer.serialize_model(model_root, self)
                self.object_store.finalize()
            else:
                model = None
                self.object_store.prepare_finalize(None)
                self.object_store.finalize()
            self._static_properties.clear()
            return model
        except Exception as e:
            LOG.exception(
                "Exception %s occurred"
                " during MuranoDslExecutor finalization", e)
            return None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finalize()



#    Copyright (c) 2015 Telefonica I+D
#    Copyright (c) 2016 AT&T Corp
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import datetime
import json
import os
import tempfile
from unittest import mock

from oslo_serialization import base64
import yaml as yamllib

from murano.common import exceptions
from murano.dsl import dsl
from murano.dsl import murano_object
from murano.dsl import murano_type
from murano.dsl import object_store
from murano.engine.system import agent
from murano.engine.system import resource_manager
from murano.tests.unit import base


class TestAgent(base.MuranoTestCase):
    def setUp(self):
        super(TestAgent, self).setUp()
        if hasattr(yamllib, 'CSafeLoader'):
            self.yaml_loader = yamllib.CSafeLoader
        else:
            self.yaml_loader = yamllib.SafeLoader

        self.override_config('disable_murano_agent', False, group='engine')
        self.override_config('signing_key', False, group='engine')

        mock_host = mock.MagicMock()
        mock_host.id = '1234'
        mock_host.find_owner = lambda *args, **kwargs: mock_host
        mock_host().getRegion.return_value = mock.MagicMock(
            __class__=dsl.MuranoObjectInterface)
        self.rabbit_mq_settings = {
            'agentRabbitMq': {'login': 'test_login',
                              'password': 'test_password',
                              'host': 'test_host',
                              'port': 123,
                              'virtual_host': 'test_virtual_host'}
        }
        mock_host().getRegion()().getConfig.return_value =\
            self.rabbit_mq_settings

        self.agent = agent.Agent(mock_host)
        self.resources = mock.Mock(spec=resource_manager.ResourceManager)
        self.resources.string.return_value = 'text'

        self.addCleanup(mock.patch.stopall)

    def _read(self, path):
        execution_plan_dir = os.path.abspath(
            os.path.join(__file__, '../execution_plans/')
        )
        with open(execution_plan_dir + "/" + path) as file:
            return file.read()

    @mock.patch('murano.common.messaging.mqclient.kombu')
    def test_send_creates_queue(self, mock_kombu):
        self.agent.send_raw({})

        # Verify that MQClient was instantiated, by checking whether
        # kombu.Connection was called.
        mock_kombu.Connection.assert_called_with(
            'amqp://{login}:{password}@{host}:{port}/{virtual_host}'.format(
                **self.rabbit_mq_settings['agentRabbitMq']
            ), ssl=None)

        # Verify that client.declare() was called by checking whether kombu
        # functions were called.
        self.assertEqual(1, mock_kombu.Exchange.call_count)
        self.assertEqual(1, mock_kombu.Queue.call_count)

    @mock.patch('murano.engine.system.agent.LOG')
    def test_send_with_murano_agent_disabled(self, mock_log):
        self.override_config('disable_murano_agent', True, group='engine')

        self.assertRaises(exceptions.PolicyViolationException,
                          self.agent.send_raw, {})

    @mock.patch('murano.engine.system.agent.Agent._sign')
    @mock.patch('murano.common.messaging.mqclient.kombu')
    def test_send(self, mock_kombu, mock_sign):
        template = yamllib.load(
            self._read('template_with_files.template'),
            Loader=self.yaml_loader)

        self.agent._queue = 'test_queue'
        mock_sign.return_value = 'SIGNATURE'
        plan = self.agent.build_execution_plan(template, self.resources)
        with mock.patch.object(self.agent, 'build_execution_plan',
                               return_value=plan):
            self.agent.send(template, self.resources)

        self.assertEqual(1, mock_kombu.Producer.call_count)
        mock_kombu.Producer().publish.assert_called_once_with(
            exchange='',
            routing_key='test_queue',
            body=json.dumps(plan),
            message_id=plan['ID'],
            headers={'signature': 'SIGNATURE'}
        )

    @mock.patch('murano.engine.system.agent.eventlet.event.Event')
    @mock.patch('murano.common.messaging.mqclient.kombu')
    def test_is_ready(self, mock_kombu, mock_event):
        v2_result = yamllib.load(
            self._read('application.template'),
            Loader=self.yaml_loader)

        mock_event().wait.side_effect = None
        mock_event().wait.return_value = v2_result

        self.assertTrue(self.agent.is_ready(1))

        mock_event().wait.side_effect = agent.eventlet.Timeout

        self.assertFalse(self.agent.is_ready(1))

    @mock.patch('murano.engine.system.agent.Agent._sign')
    @mock.patch('murano.engine.system.agent.eventlet.event.Event')
    @mock.patch('murano.common.messaging.mqclient.kombu')
    def test_call_with_v1_result(self, mock_kombu, mock_event, mock_sign):
        template = yamllib.load(
            self._read('template_with_files.template'),
            Loader=self.yaml_loader)

        test_v1_result = {
            'FormatVersion': '1.0.0',
            'IsException': False,
            'Result': [
                {
                    'IsException': False,
                    'Result': 'test_result'
                }
            ]
        }

        mock_event().wait.side_effect = None
        mock_event().wait.return_value = test_v1_result
        mock_sign.return_value = 'SIGNATURE'

        self.agent._queue = 'test_queue'
        plan = self.agent.build_execution_plan(template, self.resources)
        mock.patch.object(
            self.agent, 'build_execution_plan', return_value=plan).start()

        result = self.agent.call(template, self.resources, None)
        self.assertIsNotNone(result)
        self.assertEqual('test_result', result)

        self.assertEqual(1, mock_kombu.Producer.call_count)
        mock_kombu.Producer().publish.assert_called_once_with(
            exchange='',
            routing_key='test_queue',
            body=json.dumps(plan),
            message_id=plan['ID'],
            headers={'signature': 'SIGNATURE'}
        )

    @mock.patch('murano.engine.system.agent.Agent._sign')
    @mock.patch('murano.engine.system.agent.eventlet.event.Event')
    @mock.patch('murano.common.messaging.mqclient.kombu')
    def test_call_with_v2_result(self, mock_kombu, mock_event, mock_sign):
        template = yamllib.load(
            self._read('template_with_files.template'),
            Loader=self.yaml_loader)

        v2_result = yamllib.load(
            self._read('application.template'),
            Loader=self.yaml_loader)

        mock_event().wait.side_effect = None
        mock_event().wait.return_value = v2_result
        mock_sign.return_value = 'SIGNATURE'

        self.agent._queue = 'test_queue'
        plan = self.agent.build_execution_plan(template, self.resources)
        mock.patch.object(
            self.agent, 'build_execution_plan', return_value=plan).start()

        result = self.agent.call(template, self.resources, None)
        self.assertIsNotNone(result)
        self.assertEqual(v2_result['Body'], result)

        self.assertEqual(1, mock_kombu.Producer.call_count)
        mock_kombu.Producer().publish.assert_called_once_with(
            exchange='',
            routing_key='test_queue',
            body=json.dumps(plan),
            message_id=plan['ID'],
            headers={'signature': 'SIGNATURE'}
        )

    @mock.patch('murano.engine.system.agent.eventlet.event.Event')
    @mock.patch('murano.common.messaging.mqclient.kombu')
    def test_call_with_no_result(self, mock_kombu, mock_event):
        template = yamllib.load(
            self._read('template_with_files.template'),
            Loader=self.yaml_loader)

        mock_event().wait.side_effect = None
        mock_event().wait.return_value = None

        result = self.agent.call(template, self.resources, None)
        self.assertIsNone(result)

    @mock.patch('murano.engine.system.agent.eventlet.event.Event')
    @mock.patch('murano.common.messaging.mqclient.kombu')
    def test_call_except_timeout(self, mock_kombu, mock_event):
        self.override_config('agent_timeout', 1, group='engine')

        mock_event().wait.side_effect = agent.eventlet.Timeout

        template = yamllib.load(
            self._read('template_with_files.template'),
            Loader=self.yaml_loader)

        expected_error_msg = 'The murano-agent did not respond within 1 '\
                             'seconds'
        with self.assertRaisesRegex(exceptions.TimeoutException,
                                    expected_error_msg):
            self.agent.call(template, self.resources, None)

    @mock.patch('murano.engine.system.agent.datetime')
    def test_process_v1_result_with_error_code(self, mock_datetime):
        now = datetime.datetime.now().isoformat()
        mock_datetime.datetime.now().isoformat.return_value = now

        v1_result = {
            'IsException': True,
            'Result': [
                'Error Type',
                'Error Message',
                'Error Command',
                'Error Details'
            ]
        }

        expected_error = {
            'source': 'execution_plan',
            'command': 'Error Command',
            'details': 'Error Details',
            'message': 'Error Message',
            'type': 'Error Type',
            'timestamp': now
        }

        self.assertTrue(self._are_exceptions_equal(
            agent.AgentException, expected_error,
            self.agent._process_v1_result, v1_result))

        v1_result = {
            'IsException': False,
            'Result': [
                'Error Type',
                'Error Message',
                'Error Command',
                'Error Details',
                {
                    'IsException': True,
                    'Result': [
                        'Nested Error Type',
                        'Nested Error Message',
                        'Nested Error Command',
                        'Nested Error Details'
                    ]
                }
            ]
        }

        expected_error = {
            'source': 'command',
            'command': 'Nested Error Command',
            'details': 'Nested Error Details',
            'message': 'Nested Error Message',
            'type': 'Nested Error Type',
            'timestamp': now
        }

        self.assertTrue(self._are_exceptions_equal(
            agent.AgentException, expected_error,
            self.agent._process_v1_result, v1_result))

    def test_process_v2_result_with_error_code(self):
        v2_result = {
            'Body': {
                'Message': 'Test Error Message',
                'AdditionalInfo': 'Test Additional Info',
                'ExtraAttr': 'Test extra data'
            },
            'FormatVersion': '2.0.0',
            'Name': 'TestApp',
            'ErrorCode': 123,
            'Time': 'Right now'
        }

        expected_error = {
            'errorCode': 123,
            'message': 'Test Error Message',
            'details': 'Test Additional Info',
            'time': 'Right now',
            'extra': {'ExtraAttr': 'Test extra data'}
        }

        self.assertTrue(self._are_exceptions_equal(
            agent.AgentException, expected_error,
            self.agent._process_v2_result, v2_result))

    def _are_exceptions_equal(self, exception, expected_error, function,
                              result):
        """Checks whether expected and returned dict from exception are equal.

        Because casting dicts to strings changes the ordering of the keys,
        manual comparison of the result and expected result is performed.
        """
        try:
            # deepcopy must be performed because _process_v1_result
            # deletes attrs from the original dict passed in.
            self.assertRaises(exception, function, copy.deepcopy(result))
            function(result)
        except exception as e:
            e_string = str(e).replace("'", "\"").replace('None', 'null')
            e_dict = json.loads(e_string)
            self.assertEqual(sorted(expected_error.keys()),
                             sorted(e_dict.keys()))
            for key, val in expected_error.items():
                self.assertEqual(val, e_dict[key])
        except Exception:
            return False
        return True


class TestExecutionPlan(base.MuranoTestCase):
    def setUp(self):
        super(TestExecutionPlan, self).setUp()
        if hasattr(yamllib, 'CSafeLoader'):
            self.yaml_loader = yamllib.CSafeLoader
        else:
            self.yaml_loader = yamllib.SafeLoader

        self.override_config('signing_key', False, group='engine')

        self.mock_murano_class = mock.Mock(spec=murano_type.MuranoClass)
        self.mock_murano_class.name = 'io.murano.system.Agent'
        self.mock_murano_class.declared_parents = []
        self.mock_object_store = mock.Mock(spec=object_store.ObjectStore)

        object_interface = mock.Mock(spec=murano_object.MuranoObject)
        object_interface.id = '1234'
        object_interface.find_owner = lambda *args, **kwargs: object_interface

        self.agent = agent.Agent(object_interface)
        self.resources = mock.Mock(spec=resource_manager.ResourceManager)
        self.resources.string.return_value = 'text'
        self.uuids = ['ID1', 'ID2', 'ID3', 'ID4']
        self.mock_uuid = self._stub_uuid(self.uuids)
        time_mock = mock.patch('time.time').start()
        time_mock.return_value = 2
        self.addCleanup(mock.patch.stopall)

    def _read(self, path):
        execution_plan_dir = os.path.abspath(
            os.path.join(__file__, '../execution_plans/')
        )
        with open(execution_plan_dir + "/" + path) as file:
            return file.read()

    def test_execution_plan_v2_application_type(self):
        template = yamllib.load(
            self._read('application.template'),
            Loader=self.yaml_loader)
        template = self.agent.build_execution_plan(template, self.resources)
        self.assertEqual(self._get_application(), template)

    def test_execution_plan_v2_chef_type(self):
        template = yamllib.load(
            self._read('chef.template'),
            Loader=self.yaml_loader)
        template = self.agent.build_execution_plan(template, self.resources)
        self.assertEqual(self._get_chef(), template)

    def test_execution_plan_v2_telnet_application(self):
        template = yamllib.load(
            self._read('DeployTelnet.template'),
            Loader=self.yaml_loader)
        template = self.agent.build_execution_plan(template, self.resources)
        self.assertEqual(self._get_telnet_application(), template)

    def test_execution_plan_v2_tomcat_application(self):
        template = yamllib.load(
            self._read('DeployTomcat.template'),
            Loader=self.yaml_loader)
        template = self.agent.build_execution_plan(template, self.resources)

    def test_execution_plan_v2_app_without_files(self):
        template = yamllib.load(
            self._read('application_without_files.template'),
            Loader=self.yaml_loader)
        template = self.agent.build_execution_plan(template, self.resources)
        self.assertEqual(self._get_app_without_files(), template)

    def test_execution_plan_v2_app_with_file_in_template(self):
        template = yamllib.load(
            self._read('template_with_files.template'),
            Loader=self.yaml_loader)
        template = self.agent.build_execution_plan(template, self.resources)
        self.assertEqual(self._get_app_with_files_in_template(), template)

    def _get_application(self):
        return {
            'Action': 'Execute',
            'Body': 'return deploy(args.appName).stdout\n',
            'Files': {
                self.uuids[1]: {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'deployTomcat.sh'
                },
                self.uuids[2]: {
                    'Body': 'dGV4dA==\n',
                    'BodyType': 'Base64',
                    'Name': 'installer'
                },
                self.uuids[3]: {
                    'Body': 'dGV4dA==\n',
                    'BodyType': 'Base64',
                    'Name': 'common.sh'
                }
            },
            'FormatVersion': '2.0.0',
            'ID': self.uuids[0],
            'Stamp': 20000,
            'Name': 'Deploy Tomcat',
            'Parameters': {
                'appName': '$appName'
            },
            'Scripts': {
                'deploy': {
                    'EntryPoint': self.uuids[1],
                    'Files': [
                        self.uuids[2],
                        self.uuids[3]
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Application',
                    'Version': '1.0.0'
                }
            },
            'Version': '1.0.0'
        }

    def _get_app_with_files_in_template(self):
        return {
            'Action': 'Execute',
            'Body': 'return deploy(args.appName).stdout\n',
            'Files': {
                self.uuids[1]: {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'deployTomcat.sh'
                },
                'updateScript': {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'updateScript'
                },
            },
            'FormatVersion': '2.0.0',
            'ID': self.uuids[0],
            'Stamp': 20000,
            'Name': 'Deploy Tomcat',
            'Parameters': {
                'appName': '$appName'
            },
            'Scripts': {
                'deploy': {
                    'EntryPoint': self.uuids[1],
                    'Files': [
                        'updateScript'
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Application',
                    'Version': '1.0.0'
                }
            },
            'Version': '1.0.0'
        }

    def _get_app_without_files(self):
        return {
            'Action': 'Execute',
            'Body': 'return deploy(args.appName).stdout\n',
            'Files': {
                self.uuids[1]: {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'deployTomcat.sh'
                },
            },
            'FormatVersion': '2.0.0',
            'ID': self.uuids[0],
            'Stamp': 20000,
            'Name': 'Deploy Tomcat',
            'Parameters': {
                'appName': '$appName'
            },
            'Scripts': {
                'deploy': {
                    'EntryPoint': self.uuids[1],
                    'Files': [],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Application',
                    'Version': '1.0.0'
                }
            },
            'Version': '1.0.0'
        }

    def _get_chef(self):
        return {
            'Action': 'Execute',
            'Body': 'return deploy(args.appName).stdout\n',
            'Files': {
                self.uuids[1]: {
                    'Name': 'tomcat.git',
                    'Type': 'Downloadable',
                    'URL': 'https://github.com/tomcat.git'
                },
                self.uuids[2]: {
                    'Name': 'java',
                    'Type': 'Downloadable',
                    'URL': 'https://github.com/java.git'
                },

            },
            'FormatVersion': '2.0.0',
            'ID': self.uuids[0],
            'Stamp': 20000,
            'Name': 'Deploy Chef',
            'Parameters': {
                'appName': '$appName'
            },
            'Scripts': {
                'deploy': {
                    'EntryPoint': 'cookbook/recipe',
                    'Files': [
                        self.uuids[1],
                        self.uuids[2]
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Chef',
                    'Version': '1.0.0'
                }
            },
            'Version': '1.0.0'
        }

    def _get_telnet_application(self):
        return {
            'Action': 'Execute',
            'Body': 'return deploy(args.appName).stdout\n',
            'Files': {
                self.uuids[1]: {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'deployTelnet.sh'
                },
                self.uuids[2]: {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'installer.sh'
                },
                self.uuids[3]: {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'common.sh'
                }
            },
            'FormatVersion': '2.0.0',
            'ID': self.uuids[0],
            'Stamp': 20000,
            'Name': 'Deploy Telnet',
            'Parameters': {
                'appName': '$appName'
            },
            'Scripts': {
                'deploy': {
                    'EntryPoint': self.uuids[1],
                    'Files': [
                        self.uuids[2],
                        self.uuids[3]
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Application',
                    'Version': '1.0.0'
                }
            },
            'Version': '1.0.0'
        }

    def _stub_uuid(self, values=None):
        class FakeUUID(object):
            def __init__(self, v):
                self.hex = v

        if values is None:
            values = []
        mock_uuid4 = mock.patch('uuid.uuid4').start()
        mock_uuid4.side_effect = [FakeUUID(v) for v in values]
        return mock_uuid4

    @mock.patch('murano.engine.system.resource_manager.ResourceManager'
                '._get_package')
    def test_file_line_endings(self, _get_package):
        class FakeResources(object):
            """Class with only string() method from ResourceManager class"""
            @staticmethod
            def string(name, owner=None, binary=False):
                return resource_manager.ResourceManager.string(
                    receiver=None, name=name, owner=owner, binary=binary)

        # make path equal to provided name inside resources.string()
        package = mock.Mock()
        package.get_resource.side_effect = lambda m: m
        _get_package.return_value = package

        text = b"First line\nSecond line\rThird line\r\nFourth line"
        modified_text = u"First line\nSecond line\nThird line\nFourth line"
        encoded_text = base64.encode_as_text(text) + "\n"
        resources = FakeResources()

        with tempfile.NamedTemporaryFile() as script_file:
            script_file.write(text)
            script_file.file.flush()
            os.fsync(script_file.file.fileno())

            # check that data has been written correctly
            script_file.seek(0)
            file_data = script_file.read()
            self.assertEqual(text, file_data)

            # check resources.string() output
            # text file
            result = resources.string(script_file.name)
            self.assertEqual(modified_text, result)
            # binary file
            result = resources.string(script_file.name, binary=True)
            self.assertEqual(text, result)

            # check _get_body() output
            filename = os.path.basename(script_file.name)
            folder = os.path.dirname(script_file.name)
            # text file
            body = self.agent._get_body(filename, resources, folder)
            self.assertEqual(modified_text, body)
            # binary file
            filename = '<{0}>'.format(filename)
            body = self.agent._get_body(filename, resources, folder)
            self.assertEqual(encoded_text, body)

    def test_queue_name(self):
        self.agent._queue = 'test_queue'
        self.assertEqual(self.agent.queue_name(), self.agent._queue)

    def test_prepare_message(self):
        template = {'test'}
        msg_id = 12345
        msg = self.agent._prepare_message(template, msg_id)
        self.assertEqual(msg.id, msg_id)
        self.assertEqual(msg._body, template)

    def test_execution_plan_v1(self):
        template = yamllib.load(
            self._read('application.template'),
            Loader=self.yaml_loader)
        rtn_template = self.agent._build_v1_execution_plan(template,
                                                           self.resources)
        self.assertEqual(template, rtn_template)

    def test_get_array_item(self):
        array = [1, 2, 3]
        index = 2
        self.assertEqual(array[2], self.agent._get_array_item(array, index))

        index = 3
        self.assertIsNone(self.agent._get_array_item(array, index))

    def test_execution_plan_error(self):
        template = None
        self.assertRaises(ValueError,
                          self.agent.build_execution_plan,
                          template, self.resources)



# Copyright (c) 2015 Mirantis Inc.
# Copyright (c) 2016 AT&T Corp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslo_config import fixture as config_fixture
from oslo_serialization import jsonutils
from webob import exc

from murano.api.v1 import environments
from murano.api.v1 import sessions
from murano.services import states
import murano.tests.unit.api.base as tb
from murano.tests.unit import utils as test_utils


class TestSessionsApi(tb.ControllerTest, tb.MuranoApiTestCase):
    def setUp(self):
        super(TestSessionsApi, self).setUp()
        self.environments_controller = environments.Controller()
        self.sessions_controller = sessions.Controller()
        self.fixture = self.useFixture(config_fixture.Config())
        self.fixture.conf(args=[])

    def test_cant_deploy_from_another_tenant(self):
        """Test to prevent deployment under another tenant user's creds

        If user from one tenant uses session id and environment id
        of user from another tenant - he is not able to deploy
        the environment.

        Bug: #1382026
        """
        CREDENTIALS_1 = {'tenant': 'test_tenant_1', 'user': 'test_user_1'}
        CREDENTIALS_2 = {'tenant': 'test_tenant_2', 'user': 'test_user_2'}

        self._set_policy_rules(
            {'create_environment': '@'}
        )
        self.expect_policy_check('create_environment')

        # Create environment for user #1
        request = self._post(
            '/environments',
            jsonutils.dump_as_bytes({'name': 'test_environment_1'}),
            **CREDENTIALS_1
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)
        self.assertEqual(CREDENTIALS_1['tenant'],
                         response_body['tenant_id'])

        ENVIRONMENT_ID = response_body['id']

        # Create session of user #1
        request = self._post(
            '/environments/{environment_id}/configure'
            .format(environment_id=ENVIRONMENT_ID),
            b'',
            **CREDENTIALS_1
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)

        SESSION_ID = response_body['id']

        # Deploy the environment using environment id and session id of user #1
        # by user #2
        request = self._post(
            '/environments/{environment_id}/sessions/'
            '{session_id}/deploy'
            .format(environment_id=ENVIRONMENT_ID, session_id=SESSION_ID),
            b'',
            **CREDENTIALS_2
        )
        response = request.get_response(self.api)

        # Should be forbidden!
        self.assertEqual(403, response.status_code)

    def test_session_show(self):
        CREDENTIALS_1 = {'tenant': 'test_tenant_1', 'user': 'test_user_1'}
        CREDENTIALS_2 = {'tenant': 'test_tenant_2', 'user': 'test_user_2'}

        self._set_policy_rules(
            {'create_environment': '@'}
        )
        self.expect_policy_check('create_environment')

        # Create environment for user #1
        request = self._post(
            '/environments',
            jsonutils.dump_as_bytes({'name': 'test_environment_1'}),
            **CREDENTIALS_1
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)
        self.assertEqual(CREDENTIALS_1['tenant'],
                         response_body['tenant_id'])
        ENVIRONMENT_ID = response_body['id']

        # Create session of user #1
        request = self._post(
            '/environments/{environment_id}/configure'
            .format(environment_id=ENVIRONMENT_ID),
            b'',
            **CREDENTIALS_1
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)
        SESSION_ID = response_body['id']

        # Show environment with correct credentials
        request = self._get(
            '/environments/{environment_id}/sessions/{session_id}'
            .format(environment_id=ENVIRONMENT_ID, session_id=SESSION_ID),
            b'',
            **CREDENTIALS_1
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)
        self.assertEqual(SESSION_ID, response_body['id'])

        # Show environment with incorrect credentials
        request = self._get(
            '/environments/{environment_id}/sessions/{session_id}'
            .format(environment_id=ENVIRONMENT_ID, session_id=SESSION_ID),
            b'',
            **CREDENTIALS_2
        )
        response = request.get_response(self.api)
        self.assertEqual(403, response.status_code)

    def test_session_delete(self):
        CREDENTIALS = {'tenant': 'test_tenant_1', 'user': 'test_user_1'}

        self._set_policy_rules(
            {'create_environment': '@'}
        )
        self.expect_policy_check('create_environment')

        # Create environment
        request = self._post(
            '/environments',
            jsonutils.dump_as_bytes({'name': 'test_environment_1'}),
            **CREDENTIALS
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)
        self.assertEqual(CREDENTIALS['tenant'],
                         response_body['tenant_id'])
        ENVIRONMENT_ID = response_body['id']

        # Create session
        request = self._post(
            '/environments/{environment_id}/configure'
            .format(environment_id=ENVIRONMENT_ID),
            b'',
            **CREDENTIALS
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)
        SESSION_ID = response_body['id']

        # Delete session
        request = self._delete(
            '/environments/{environment_id}/delete/{session_id}'
            .format(environment_id=ENVIRONMENT_ID, session_id=SESSION_ID),
            b'',
            **CREDENTIALS
        )
        response = self.sessions_controller.delete(
            request, ENVIRONMENT_ID, SESSION_ID)

        # Make sure the session was deleted
        request = self._get(
            '/environments/{environment_id}/sessions/{session_id}'
            .format(environment_id=ENVIRONMENT_ID, session_id=SESSION_ID),
            b'',
            **CREDENTIALS
        )
        response = request.get_response(self.api)
        self.assertEqual(404, response.status_code)

    @mock.patch('murano.api.v1.sessions.sessions.SessionServices')
    @mock.patch('murano.api.v1.sessions.envs')
    @mock.patch('murano.api.v1.sessions.check_env')
    def test_configure(self, _, mock_envs, mock_session_services):
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        mock_session = mock.MagicMock(to_dict=mock.MagicMock(
            return_value={'test_env_id', 'test_user_id'}))
        mock_session_services.create.return_value = mock_session

        result = self.sessions_controller.configure(
            mock_request, 'test_env_id')
        self.assertEqual({'test_env_id', 'test_user_id'}, result)

    @mock.patch('murano.api.v1.sessions.envs')
    @mock.patch('murano.api.v1.sessions.check_env')
    def test_configure_with_env_in_illegal_state(self, _, mock_envs):
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        illegal_states = [states.EnvironmentStatus.DEPLOYING,
                          states.EnvironmentStatus.DELETING]
        expected_error_msg = 'Could not open session for environment '\
                             '<EnvId: {env_id}>, environment has '\
                             'deploying or deleting status.'\
                             .format(env_id='test_env_id')

        for state in illegal_states:
            mock_envs.EnvironmentServices.get_status.return_value = state

            with self.assertRaisesRegex(exc.HTTPForbidden,
                                        expected_error_msg):
                self.sessions_controller.configure(mock_request, 'test_env_id')

    @mock.patch('murano.api.v1.sessions.check_session')
    @mock.patch('murano.api.v1.sessions.db_session')
    def test_show_with_invalid_user(self, mock_db_session, _):
        mock_session = mock.MagicMock(user_id='test_user_id')
        mock_db_session.get_session().query().get.return_value = mock_session
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        mock_request.context.user = 'another_test_user_id'
        expected_error_msg = 'User <UserId {0}> is not authorized to '\
            'access session <SessionId {1}>.'\
            .format(mock_request.context.user, 'test_sess_id')

        with self.assertRaisesRegex(exc.HTTPUnauthorized,
                                    expected_error_msg):
            self.sessions_controller.show(mock_request, None, 'test_sess_id')

    @mock.patch('murano.api.v1.sessions.check_session')
    @mock.patch('murano.api.v1.sessions.sessions.SessionServices')
    @mock.patch('murano.api.v1.sessions.db_session')
    def test_show_with_invalid_session(self, mock_db_session,
                                       mock_session_services, _):
        mock_session = mock.MagicMock(user_id='test_user_id')
        mock_db_session.get_session().query().get.return_value = mock_session
        mock_session_services.validate.return_value = False
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        mock_request.context.user = 'test_user_id'
        expected_error_msg = 'Session <SessionId {0}> is invalid: environment'\
                             ' has been updated or updating right now with'\
                             ' other session'.format('test_sess_id')

        with self.assertRaisesRegex(exc.HTTPForbidden,
                                    expected_error_msg):
            self.sessions_controller.show(mock_request, None, 'test_sess_id')

    @mock.patch('murano.api.v1.sessions.check_session')
    @mock.patch('murano.api.v1.sessions.db_session')
    def test_delete_with_invalid_user(self, mock_db_session, _):
        mock_session = mock.MagicMock(user_id='test_user_id')
        mock_db_session.get_session().query().get.return_value = mock_session
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        mock_request.context.user = 'another_test_user_id'
        expected_error_msg = 'User <UserId {0}> is not authorized to '\
            'access session <SessionId {1}>.'\
            .format(mock_request.context.user, 'test_sess_id')

        with self.assertRaisesRegex(exc.HTTPUnauthorized,
                                    expected_error_msg):
            self.sessions_controller.delete(mock_request, None, 'test_sess_id')

    @mock.patch('murano.api.v1.sessions.check_session')
    @mock.patch('murano.api.v1.sessions.sessions.SessionServices')
    @mock.patch('murano.api.v1.sessions.db_session')
    def test_delete_with_deploying_session(self, mock_db_session,
                                           mock_session_services, _):
        mock_session = mock.MagicMock(user_id='test_user_id',
                                      state=states.SessionState.DEPLOYING)
        mock_db_session.get_session().query().get.return_value = mock_session
        mock_session_services.validate.return_value = False
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        mock_request.context.user = 'test_user_id'
        expected_error_msg = 'Session <SessionId: {0}> is in deploying '\
                             'state and could not be deleted'\
                             .format('test_sess_id')

        with self.assertRaisesRegex(exc.HTTPForbidden,
                                    expected_error_msg):
            self.sessions_controller.delete(mock_request, None, 'test_sess_id')

    @mock.patch('murano.api.v1.sessions.check_session')
    @mock.patch('murano.api.v1.sessions.sessions.SessionServices')
    @mock.patch('murano.api.v1.sessions.db_session')
    def test_deploy_with_invalid_session(self, mock_db_session,
                                         mock_session_services, _):
        mock_db_session.get_session().query().get.return_value = None
        mock_session_services.validate.return_value = False
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        expected_error_msg = 'Session <SessionId {0}> is invalid: environment'\
                             ' has been updated or updating right now with'\
                             ' other session'.format('test_sess_id')

        with self.assertRaisesRegex(exc.HTTPForbidden,
                                    expected_error_msg):
            self.sessions_controller.deploy(mock_request, None, 'test_sess_id')

    @mock.patch('murano.api.v1.sessions.check_session')
    @mock.patch('murano.api.v1.sessions.sessions.SessionServices')
    @mock.patch('murano.api.v1.sessions.db_session')
    def test_deploy_with_session_in_invalid_state(self, mock_db_session,
                                                  mock_session_services, _):
        mock_session_services.validate.return_value = True
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        expected_error_msg = 'Session <SessionId {0}> is already deployed or '\
                             'deployment is in progress'.format('test_sess_id')

        invalid_states = [states.SessionState.DEPLOYING,
                          states.SessionState.DEPLOYED,
                          states.SessionState.DEPLOY_FAILURE,
                          states.SessionState.DELETING,
                          states.SessionState.DELETE_FAILURE]

        for state in invalid_states:
            mock_session = mock.MagicMock(state=state)
            mock_db_session.get_session().query().get.return_value =\
                mock_session
            with self.assertRaisesRegex(exc.HTTPForbidden,
                                        expected_error_msg):
                self.sessions_controller.deploy(
                    mock_request, None, 'test_sess_id')



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
  *** Deprecation warning ***
  This file is about to be deprecated, please use python-muranoclient.
  *** Deprecation warning ***
"""
import sys
import traceback

from oslo_config import cfg
from oslo_db import exception as db_exception
from oslo_log import log as logging

from murano.common import consts
from murano.db.catalog import api as db_catalog_api
from murano.packages import load_utils
from murano import version

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class AdminContext(object):
    def __init__(self):
        self.is_admin = True


def _do_import_package(_dir, categories, update=False):
    LOG.debug("Going to import Murano package from {source}".format(
        source=_dir))
    pkg = load_utils.load_from_dir(_dir)

    LOG.debug("Checking for existing packages")
    existing = db_catalog_api.package_search(
        {'fqn': pkg.full_name},
        AdminContext())
    if existing:
        existing_pkg = existing[0]
        if update:
            LOG.debug('Deleting existing package {exst_pkg_id}').format(
                exst_pkg_id=existing_pkg.id)
            db_catalog_api.package_delete(existing_pkg.id, AdminContext())
        else:
            LOG.error("Package '{name}' exists ({pkg_id}). Use --update."
                      .format(name=pkg.full_name, pkg_id=existing_pkg.id))
            return

    package = {
        'fully_qualified_name': pkg.full_name,
        'type': pkg.package_type,
        'author': pkg.author,
        'supplier': pkg.supplier,
        'name': pkg.display_name,
        'description': pkg.description,
        # note: we explicitly mark all the imported packages as public,
        # until a parameter added to control visibility scope of a package
        'is_public': True,
        'tags': pkg.tags,
        'logo': pkg.logo,
        'supplier_logo': pkg.supplier_logo,
        'ui_definition': pkg.ui,
        'class_definitions': pkg.classes,
        'archive': pkg.blob,
        'categories': categories or []
    }

    # note(ruhe): the second parameter is tenant_id
    # it is a required field in the DB, that's why we pass an empty string
    result = db_catalog_api.package_upload(package, '')

    LOG.info("Finished import of package {res_id}".format(res_id=result.id))


# TODO(ruhe): proper error handling
def do_import_package():
    """Import Murano package from local directory."""
    _do_import_package(
        CONF.command.directory,
        CONF.command.categories,
        CONF.command.update)


def do_list_categories():
    categories = db_catalog_api.category_get_names()

    if categories:
        print(">> Murano package categories:")
        for c in categories:
            print("* {0}".format(c))
    else:
        print("No categories were found")


def do_add_category():
    category_name = CONF.command.category_name

    try:
        db_catalog_api.category_add(category_name)
        print(">> Successfully added category {0}".format(category_name))
    except db_exception.DBDuplicateEntry:
        print(">> ERROR: Category '{0}' already exists".format(category_name))


def add_command_parsers(subparsers):
    parser = subparsers.add_parser('import-package')
    parser.set_defaults(func=do_import_package)
    parser.add_argument('directory',
                        help='A directory with Murano package.')
    parser.add_argument('-u', '--update',
                        action="store_true",
                        default=False,
                        help='If a package already exists, delete and update')
    parser.add_argument('-c', '--categories',
                        choices=consts.CATEGORIES,
                        nargs='*',
                        help='An optional list of categories this package '
                             'to be assigned to.')

    parser = subparsers.add_parser('category-list')
    parser.set_defaults(func=do_list_categories)

    parser = subparsers.add_parser('category-add')
    parser.set_defaults(func=do_add_category)
    parser.add_argument('category_name',
                        help='Name of the new category.')


command_opt = cfg.SubCommandOpt('command',
                                title='Commands',
                                help='Show available commands.',
                                handler=add_command_parsers)


def main():
    CONF.register_cli_opt(command_opt)

    try:
        default_config_files = cfg.find_config_files('murano', 'murano')
        CONF(sys.argv[1:], project='murano', prog='murano-manage',
             version=version.version_string,
             default_config_files=default_config_files)
    except RuntimeError as e:
        LOG.error("failed to initialize murano-manage: {error}".format(
            error=e))
        sys.exit("ERROR: %s" % e)

    try:
        CONF.command.func()
    except Exception as e:
        tb = traceback.format_exc()
        err_msg = ("murano-manage command failed: {error}\n"
                   "{traceback}").format(error=e, traceback=tb)
        LOG.error(err_msg)
        sys.exit(err_msg)


if __name__ == '__main__':
    main()



# Copyright 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime as dt
import time
from unittest import mock

from oslo_utils import timeutils

from murano.common import statservice
from murano.db import models
from murano.db import session as db_session
from murano.services import states
from murano.tests.unit import base


@mock.patch('murano.common.statservice.v1.stats',
            request_count=11, error_count=22, average_time=33,
            requests_per_tenant=44)
class StatsCollectingServiceTest(base.MuranoTestCase):

    def setUp(self):
        super(StatsCollectingServiceTest, self).setUp()
        self.service = statservice.StatsCollectingService()
        self.service._prev_time = 0
        self.mock_new_stats = mock.MagicMock(request_count=1, error_count=2)
        self.service._stats_db = mock.MagicMock(
            get_stats_by_host=mock.MagicMock(return_value=self.mock_new_stats))
        statservice.LOG = mock.MagicMock()

    def test_service_start_and_stop(self, _):
        self.assertEqual(0, len(self.service.tg.threads))
        self.service.start()
        self.assertEqual(2, len(self.service.tg.threads))
        self.service.stop()
        self.assertEqual(0, len(self.service.tg.threads))

    @mock.patch('murano.common.statservice.time')
    def test_update_stats(self, mock_time, mock_stats):
        now = time.time()
        mock_time.time.return_value = now

        self.service.update_stats()

        statservice.LOG.debug.assert_any_call(
            'Stats: (Requests: 11  Errors: 22  Ave.Res.Time 33.0000\n Per '
            'tenant: 44)')
        self.assertEqual(now, self.service._prev_time)
        self.assertEqual(mock_stats.request_count,
                         self.mock_new_stats.request_count)
        self.assertEqual(mock_stats.error_count,
                         self.mock_new_stats.error_count)
        self.assertEqual(mock_stats.average_time,
                         self.mock_new_stats.average_response_time)
        self.assertEqual(str(mock_stats.requests_per_tenant),
                         self.mock_new_stats.requests_per_tenant)
        self.assertEqual((11 - 1) / now,
                         self.mock_new_stats.requests_per_second)
        self.assertEqual((22 - 2) / now,
                         self.mock_new_stats.errors_per_second)
        self.service._stats_db.update.assert_called_once_with(
            self.service._hostname, self.mock_new_stats)

    @mock.patch('murano.common.statservice.multiprocessing.cpu_count',
                return_value=5)
    @mock.patch('murano.common.statservice.psutil.cpu_percent',
                return_value=12.3)
    def test_update_stats_with_create_stats_db(self, _, __, mock_stats):
        self.service._stats_db.get_stats_by_host.return_value = None

        result = self.service.update_stats()

        self.assertIsNone(result)
        self.service._stats_db.create.assert_called_once_with(
            self.service._hostname, mock_stats.request_count,
            mock_stats.error_count, mock_stats.average_time,
            mock_stats.requests_per_tenant, 5, 12.3
        )

    @mock.patch('murano.common.statservice.LOG')
    def test_update_stats_handle_exception(self, mock_log, _):
        self.service._stats_db.update.side_effect =\
            Exception('test_error_code')

        self.service.update_stats()

        mock_log.exception.assert_called_once_with(
            "Failed to get statistics object from a "
            "database. {error_code}".format(error_code='test_error_code'))


class EnvReportingTest(base.MuranoNotifyWithDBTestCase):

    def setUp(self):
        super(EnvReportingTest, self).setUp()
        self.service = statservice.StatsCollectingService()

    @mock.patch('murano.common.statservice.status_reporter.'
                'Notification.report')
    def test_report_env_stats(self, mock_notifier):
        now = timeutils.utcnow()
        later = now + dt.timedelta(minutes=1)

        session = db_session.get_session()

        environment1 = models.Environment(
            name='test_environment1', tenant_id='test_tenant_id1',
            version=2, id='test_env_id_1',
            created=now,
            updated=later,
            description={
                'Objects': {
                    'applications': ['app1'],
                    'services': ['service1']
                }
            }
        )
        environment2 = models.Environment(
            name='test_environment2', tenant_id='test_tenant_id2',
            version=1, id='test_env_id_2',
            created=now,
            updated=later,
            description={
                'Objects': {
                    'applications': ['app2'],
                    'services': ['service3']
                }
            }
        )
        environment3 = models.Environment(
            name='test_environment3', tenant_id='test_tenant_id2',
            version=1, id='test_env_id_3',
            created=now,
            updated=later,
            description={}
        )

        session_1 = models.Session(
            environment=environment1, user_id='test_user_id',
            description={},
            state=states.SessionState.DEPLOYED,
            version=1
        )

        session_2 = models.Session(
            environment=environment2, user_id='test_user_id',
            description={},
            state=states.SessionState.DEPLOYED,
            version=0
        )

        session_3 = models.Session(
            environment=environment3, user_id='test_user_id',
            description={},
            state=states.SessionState.DEPLOY_FAILURE,
            version=1
        )

        task_1 = models.Task(
            id='task_id_1',
            environment=environment1,
            description={},
            created=now,
            started=now,
            updated=later,
            finished=later
        )

        task_2 = models.Task(
            id='task_id_2',
            environment=environment2,
            description={},
            created=now,
            started=now,
            updated=later,
            finished=later
        )

        task_3 = models.Task(
            id='task_id_3',
            environment=environment3,
            description={},
            created=now,
            started=now,
            updated=later,
            finished=later
        )

        status_1 = models.Status(
            id='status_id_1',
            task_id='task_id_1',
            text='Deployed',
            level='info'
        )

        status_2 = models.Status(
            id='status_id_2',
            task_id='task_id_2',
            text='Deployed',
            level='info'
        )

        status_3 = models.Status(
            id='status_id_3',
            task_id='task_id_3',
            text='Something was wrong',
            level='error'
        )

        session.add_all([environment1, environment2, environment3])
        session.add_all([session_1, session_2, session_3])
        session.add_all([task_1, task_2, task_3])
        session.add_all([status_1, status_2, status_3])

        session.flush()

        self.service.report_env_stats()

        self.assertEqual(mock_notifier.call_count, 2)

        dict_env_1 = {'version': 2,
                      'updated': later,
                      'tenant_id': u'test_tenant_id1',
                      'created': now,
                      'description_text': u'',
                      'status': 'ready',
                      'id': u'test_env_id_1',
                      'name': u'test_environment1'}

        dict_env_2 = {'version': 1,
                      'updated': later,
                      'tenant_id': u'test_tenant_id2',
                      'created': now,
                      'description_text': u'',
                      'status': 'ready',
                      'id': u'test_env_id_2',
                      'name': u'test_environment2'}

        calls = [mock.call('environment.exists', dict_env_1),
                 mock.call('environment.exists', dict_env_2)]

        mock_notifier.assert_has_calls(calls)



# Copyright (c) 2016 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glanceclient.v2.client as gclient
from oslo_config import cfg

from murano.common import auth_utils
from murano.dsl import dsl
from murano.dsl import helpers
from murano.dsl import session_local_storage


CONF = cfg.CONF


@dsl.name('io.murano.system.MetadefBrowser')
class MetadefBrowser(object):
    def __init__(self, this, region_name=None, cache=True):
        session = helpers.get_execution_session()
        self._project_id = session.project_id
        self._region = this.find_owner('io.murano.CloudRegion')
        self._region_name = region_name
        self._cache = cache
        self._namespaces = {}
        self._objects = {}

    @staticmethod
    @session_local_storage.execution_session_memoize
    def _get_client(region_name):
        return gclient.Client(**auth_utils.get_session_client_parameters(
            service_type='image', region=region_name, conf='glance'
        ))

    @property
    def _client(self):
        region = self._region_name or (
            None if self._region is None else self._region['name'])
        return self._get_client(region)

    def get_namespaces(self, resource_type):
        if not self._cache or resource_type not in self._namespaces:
            nss = list(self._client.metadefs_namespace.list(
                resource_type=resource_type))
            self._namespaces[resource_type] = nss
            return nss
        else:
            return self._namespaces[resource_type]

    def get_objects(self, namespace):
        if not self._cache or namespace not in self._objects:
            objects = list(self._client.metadefs_object.list(
                namespace=namespace))
            self._objects[namespace] = objects
            return objects
        else:
            return self._objects[namespace]



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import weakref


class ClassUsages(object):
    Class = 'Class'
    Meta = 'Meta'
    All = {Class, Meta}


class MetaCardinality(object):
    One = 'One'
    Many = 'Many'
    All = {One, Many}


class MetaTargets(object):
    Package = 'Package'
    Type = 'Type'
    Property = 'Property'
    Method = 'Method'
    Argument = 'Argument'
    All = {Package, Type, Property, Method, Argument}


class PropertyUsages(object):
    In = 'In'
    Out = 'Out'
    InOut = 'InOut'
    Runtime = 'Runtime'
    Const = 'Const'
    Config = 'Config'
    Static = 'Static'
    All = {In, Out, InOut, Runtime, Const, Config, Static}
    Writable = {Out, InOut, Runtime, Static, Config}


class MethodUsages(object):
    Action = 'Action'
    Runtime = 'Runtime'
    Static = 'Static'
    Extension = 'Extension'

    All = {Action, Runtime, Static, Extension}
    InstanceMethods = {Runtime, Action}
    StaticMethods = {Static, Extension}


class MethodScopes(object):
    Session = 'Session'
    Public = 'Public'

    All = {Session, Public}


class MethodArgumentUsages(object):
    Standard = 'Standard'
    VarArgs = 'VarArgs'
    KwArgs = 'KwArgs'
    All = {Standard, VarArgs, KwArgs}


class DumpTypes(object):
    Serializable = 'Serializable'
    Inline = 'Inline'
    Mixed = 'Mixed'
    All = {Serializable, Inline, Mixed}


class MuranoType(object):
    pass


class MuranoClass(MuranoType):
    pass


class MuranoMetaClass(MuranoClass):
    pass


class MuranoObject(object):
    pass


class MuranoMethod(object):
    pass


class MuranoMethodArgument(object):
    pass


class MuranoPackage(object):
    pass


class MuranoProperty(object):
    pass


class MuranoTypeReference(object):
    def __init__(self, murano_type):
        self.__murano_type = weakref.ref(murano_type)

    @property
    def type(self):
        return self.__murano_type()

    def __repr__(self):
        return '*' + repr(self.type)

    def __eq__(self, other):
        if not isinstance(other, MuranoTypeReference):
            return False
        return self.type == other.type

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.type)


class YaqlExpression(object):
    pass


class MuranoObjectInterface(object):
    pass


class ExpressionFilePosition(object):
    def __init__(self, file_path, start_line, start_column,
                 end_line, end_column):
        self._file_path = file_path
        self._start_line = start_line
        self._start_column = start_column
        self._end_line = end_line
        self._end_column = end_column

    @property
    def file_path(self):
        return self._file_path

    @property
    def start_line(self):
        return self._start_line

    @property
    def start_column(self):
        return self._start_column

    @property
    def end_line(self):
        return self._end_line

    @property
    def end_column(self):
        return self._end_column
